package com.ringlord;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.ringlord.boxer2.Builder;
import com.ringlord.boxer2.CLI;
import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Util;


/**
 * Launches the CLI; if the CLI has no work to do (no parameters), it launches
 * the Java 8 JavaFX GUI; if the Java 8 JavaFX GUI is unavailable, it launches
 * the Java 7 Swing GUI; if the Java 7 Swing GUI is unavailable, it crawls off
 * into a corner to cry bitter tears.
 *
 * @author K Udo Schuermann
 */
public class Boxer
{
  private static final boolean NO_DEFAULT_JAVAFX8 = false;


  public static void main( final String[] args )
  {
    final Logger logger = Util.logger();

    final String NL = System.getProperty( "line.separator" );
    final Runtime rt = Runtime.getRuntime();
    final StringBuilder sb = new StringBuilder();
    sb.append( "Starting..." + NL )
      .append( "\tBoxer " + Util.VERSION + " (" + Util.VERDATE + ")" + NL )
      .append( "\tCopyright \u00a9 " + Util.CPYEARS + " by K. Udo Schuermann" + NL )
      .append( "\tAll rights reserved" + NL )
      .append( NL )
      .append( "Environment:" + NL )
      .append( "\tJava " )
      .append( System.getProperty( "java.runtime.version" ) )
      .append( " (" )
      .append( System.getProperty( "java.vm.name" ) )
      .append( " " )
      .append( System.getProperty( "java.vm.version" ) )
      .append( ")" + NL + "\t" )
      .append( System.getProperty( "os.name" ) )
      .append( " " )
      .append( System.getProperty( "os.version" ) )
      .append( " (" )
      .append( System.getProperty( "os.arch" ) )
      .append( ")" )
      .append( NL + "\tAvailable processors: " )
      .append( rt.availableProcessors() )
      .append( NL + "\tMemory: " )
      .append( String.format( "%,1.2fM of %,1.2fM",
                              rt.freeMemory() / (1024.0d * 1024.0d),
                              rt.maxMemory() / (1024.0d * 1024.0d) ) )
      .append( NL + "\tCountry_language: " )
      .append( System.getProperty( "user.country" ) )
      .append( "_" )
      .append( System.getProperty( "user.language" ) )
      .append( NL + "\tDefault file encoding: " )
      .append( System.getProperty( "file.encoding" ) )
      .append( NL + "\tGraphics environment: " )
      .append( System.getProperty( "java.awt.graphicsenv" ) );

    logger.log( Level.CONFIG,
                sb.toString() );

    // Launch the Thread that verifies that the JarSigner binary is
    // available.
    Builder.verifyJarSignerBinaryIsAvailable();

    final Map<String,Object> guiParams = (args.length == 0
        ? new HashMap<String,Object>()
        : CLI.go( args ));

    if( guiParams != null )
      {
        final String guiStyle = (String)guiParams.get( "forceGUI" );
        @SuppressWarnings("unchecked") final List<Project> projects = (List<Project>)guiParams.get( "projects" );

        if( "7".equals( guiStyle ) )
          {
            // Try only Java Swing (standard through Java 7)
            try
              {
                logger.log( Level.CONFIG,
                            "Forcing Java 7 Swing GUI" );
                com.ringlord.boxer2.swing7.Launch.go( projects );
              }
            catch( final Throwable noSwing7 )
              {
                System.err.println( "Sorry, Swing 7 Swing GUI is not available." );
                logger.log( Level.SEVERE,
                            "Swing GUI not available",
                            noSwing7 );
              }
          }
        else if( "8".equals( guiStyle ) )
          {
            // Try only Java 8 JavaFX (new/preferred since Java 8)
            try
              {
                logger.log( Level.CONFIG,
                            "Forcing JavaFX 8 GUI" );
                com.ringlord.boxer2.javafx8.Launch.go( projects );
              }
            catch( final Throwable noJavaFX8 )
              {
                System.err.println( "Sorry, JavaFX 8 GUI is not accessible." );
                logger.log( Level.SEVERE,
                            "JavaFX 8 GUI not available",
                            noJavaFX8 );
              }
          }
        else
          {
            // Try Java 8 JavaFX first, then fall back to Swing
            if( NO_DEFAULT_JAVAFX8 )
              {
                try
                  {
                    logger.log( Level.FINE,
                                "Trying JavaFX 8 GUI" );
                    com.ringlord.boxer2.javafx8.Launch.go( projects );
                  }
                catch( final Throwable noJavaFX8 )
                  {
                    try
                      {
                        logger.log( Level.WARNING,
                                    "JavaFX 8 failed, trying Swing GUI" );
                        com.ringlord.boxer2.swing7.Launch.go( projects );
                      }
                    catch( final Throwable noSwing7 )
                      {
                        System.err.println( "Neither JavaFX 8 nor Java 7 Swing GUI available." );
                        logger.log( Level.SEVERE,
                                    "Java 7 Swing GUI failed, no JavaFX 8 available, either." +
                                        NL +
                                        "Try --help parameter for CLI usage." );
                      }
                  }
              }
            else
              {
                try
                  {
                    com.ringlord.boxer2.swing7.Launch.go( projects );
                  }
                catch( final Throwable noSwing7 )
                  {
                    System.err.println( "Java 7 Swing GUI is not available." );
                    logger.log( Level.SEVERE,
                                "Java 7 Swing GUI failed." + NL + "Try --help parameter for CLI usage." );
                  }
              }
          }
      }
  }
}
