package com.ringlord.boxer2;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.List;
import java.util.logging.Level;


/**
 * A utility class to collect all the entry points, i.e. methods with a
 * signature "<code>public&nbsp;static&nbsp;void&nbsp;main(String[])</code>".
 *
 * @author K Udo Schuermann
 */
public class EntryPointCollector
{
  public static void collect( final List<Class<?>> javaEntryPoints,
                              final File path )
  {
    try
      {
        // NOTE: The trailing / is important, otherwise the URL is treated as a
        // file rather than a path:
        final URL[] urls = {new URL( "file://" + path + "/" )};
        final ClassLoader cl = URLClassLoader.newInstance( urls );

        collect( javaEntryPoints,
                 path.toString(),
                 path.listFiles(),
                 cl );
      }
    catch( final MalformedURLException x )
      {
        Util.logger().log( Level.SEVERE,
                           "Cannot collect Java Application entry points from file://" + path + "/",
                           x );
      }
  }


  private static void collect( final List<Class<?>> javaEntryPoints,
                               final String rootName,
                               final File[] files,
                               final ClassLoader cl )
  {
    if( files == null )
      {
        return;
      }
    for( final File f : files )
      {
        if( f.isDirectory() )
          {
            collect( javaEntryPoints,
                     rootName,
                     f.listFiles(),
                     cl );
          }
        else if( f.isFile() )
          {
            String name = f.toString();
            if( name.endsWith( ".class" ) )
              {
                if( name.startsWith( rootName ) )
                  {
                    name = name.substring( rootName.length() );
                  }
                if( name.startsWith( "/" ) || name.startsWith( "\\" ) )
                  {
                    name = name.substring( 1 );
                  }
                // -6 because of ".class"
                String className = name.substring( 0,
                                                   name.length() - 6 );
                className = className.replace( '/',
                                               '.' ).replace( '\\',
                                                              '.' );
                try
                  {
                    final Class<?> c = cl.loadClass( className );
                    javaEntryPoints.add( c );
                  }
                catch( final ClassNotFoundException x )
                  {
                    Util.logger().log( Level.SEVERE,
                                       "ERROR: Cannot load project's class from " + rootName + ": " + className,
                                       x );
                  }
                catch( final NoClassDefFoundError x )
                  {
                    Util.logger().log( Level.SEVERE,
                                       "ERROR: Cannot load project's class from " + rootName + ": " + className,
                                       x );
                  }
                catch( final SecurityException x )
                  {
                    // Class is not public, cannot access method
                  }
              }
          }
      }
  }


  private EntryPointCollector()
  {
  }
}
