package com.ringlord.boxer2;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import com.ringlord.archive.TapeArchive;


/**
 * This is used for the .tar as well as .tar.gz/.tgz distribution archives.
 *
 * @author K Udo Schuermann
 */
public class PackageWriterTar
  implements
    PackageWriter
{
  public PackageWriterTar( final File filename )
    throws FileNotFoundException
  {
    this.tar = new TapeArchive.Writer( filename );
  }


  @Override
  public void close()
    throws IOException
  {
    tar.close();
  }


  @Override
  public void add( final File file,
                   final String name )
    throws IOException
  {
    try( final InputStream f = new BufferedInputStream( new FileInputStream( file ) ) )
      {
        add( f,
             name,
             file.lastModified() );
      }
  }


  @Override
  public void add( final InputStream input,
                   final String name,
                   final long fileTime )
    throws IOException
  {
    final TapeArchive.Entry entry = tar.addEntry( name.replace( '\\',
                                                                '/' ) );
    entry.setMTime( fileTime );

    int inBuffer;
    while( (inBuffer = input.read( BUFFER )) > -1 )
      {
        tar.write( BUFFER,
                   0,
                   inBuffer );
      }
  }

  final TapeArchive.Writer tar;
  final byte[] BUFFER = new byte[4096];
}
