package com.ringlord.boxer2;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.Adler32;
import java.util.zip.CheckedOutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;


/**
 * This is used for the JAR as well as a .zip distribution archive.
 *
 * @author K Udo Schuermann
 */
public class PackageWriterZip
  implements
    PackageWriter
{
  public PackageWriterZip( final File filename )
    throws IOException
  {
    super();
    this.zip = new ZipOutputStream( new CheckedOutputStream( new FileOutputStream( filename ),
                                                             new Adler32() ) );
    this.zip.setMethod( ZipOutputStream.DEFLATED );
    this.zip.setLevel( 9 );
  }


  @Override
  public void close()
    throws IOException
  {
    zip.finish();
    zip.close();
  }


  @Override
  public void add( final File file,
                   final String name )
    throws IOException
  {
    try( final InputStream f = new BufferedInputStream( new FileInputStream( file ) ) )
      {
        add( f,
             name,
             file.lastModified() );
      }
  }


  @Override
  public void add( final InputStream f,
                   final String name,
                   final long fileTime )
    throws IOException
  {
    final ZipEntry e = new ZipEntry( name.replace( '\\',
                                                   '/' ) );
    e.setMethod( ZipEntry.DEFLATED );
    e.setTime( fileTime );
    zip.putNextEntry( e );

    Util.copy( f,
               zip,
               BUFFER );

    zip.flush();
    zip.closeEntry();
  }

  private final ZipOutputStream zip;
  private final byte[] BUFFER = new byte[4096];
}
