package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.BorderFactory;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;

import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Project.Addition;
import com.ringlord.boxer2.Project.Addition.Where;
import com.ringlord.swing.FilePicker;
import com.ringlord.swing.Form;


public class ConfigAdditions
  extends ConfigurationPanel
{
  public ConfigAdditions( final Project project )
  {
    super();

    final Form form = new Form( new Insets( 4,
                                            8,
                                            4,
                                            0 ) );

    form.setBorder( BorderFactory.createEmptyBorder( 8,
                                                     8,
                                                     8,
                                                     8 ) );
    form.setName( "ConfigAdditions" );

    // ==================================================

    final RowEditor additions = new RowEditor( new AdditionsModel( project ),
                                               "Add file entry" );
    form.addRow( additions );

    final JLabel separator = new JLabel( "JAR File Manifest Additions:" );
    separator.setBorder( BorderFactory.createEmptyBorder( 0,
                                                          4,
                                                          4,
                                                          0 ) );
    final JTextArea manifestAdditions = new JTextArea();
    manifestAdditions.setLineWrap( false );
    manifestAdditions.setWrapStyleWord( false );
    manifestAdditions.setFont( new Font( "MonoSpaced",
                                         Font.PLAIN,
                                         manifestAdditions.getFont().getSize() ) );
    final JPanel manifestAdditionsPane = new JPanel( new BorderLayout() );
    manifestAdditionsPane.add( BorderLayout.NORTH,
                               separator );
    manifestAdditionsPane.add( BorderLayout.CENTER,
                               new JScrollPane( manifestAdditions ) );
    manifestAdditionsPane.setBorder( BorderFactory.createEmptyBorder( 12,
                                                                      12,
                                                                      8,
                                                                      12 ) );

    boolean isFirst = true;
    for( final String line : project.getManifestAdditions() )
      {
        if( isFirst )
          {
            isFirst = false;
          }
        else
          {
            manifestAdditions.append( "\n" );
          }
        manifestAdditions.append( line );
      }
    setToolTipText( "<html>"
                        + "Additions to the JAR's manifest must be in 'key: value' format. The<br>"
                        + "following are some possible keys, but you should understand their<br>"
                        + "purpose before you actually use them:<br>"
                        + "Application-Name, Permissions, Codebase, Trusted-Only,<br>"
                        + "Specification-Title, Specification-Version, Specification-Vendor,<br>"
                        + "Implementation-Title, Implementation-Version, Implementation-Vendor.<br>"
                        + "<br>"
                        + "The value portions resolves ${VERSION}, ${OS}, ${USER} and ${BUILD}",
                    separator,
                    manifestAdditions );
    manifestAdditions.addFocusListener( new FocusListener()
    {
      @Override
      public void focusGained( final FocusEvent arg0 )
      {
      }


      @Override
      public void focusLost( final FocusEvent arg0 )
      {
        final String lines = manifestAdditions.getText();
        final StringTokenizer t = new StringTokenizer( lines,
                                                       "\r\n" );
        final List<String> additions = new ArrayList<>();
        while( t.hasMoreTokens() )
          {
            final String line = t.nextToken().trim();
            if( line.length() > 0 ) // don't trim,
              {
                additions.add( line );
              }
          }
        project.setManifestAdditions( additions );
      }
    } );

    // ==================================================

    final JTextArea info = new JTextArea( 3,
                                          40 );
    info.setEditable( false );
    info.setLineWrap( true );
    info.setWrapStyleWord( true );
    info.setBackground( getBackground() );
    info.setText( "Files that you have not specified as part of your project, but still want/need to "
                  + "distribute are specified here. This might include documentation, README and "
                  + "license files, as well as sample databases, and anything else external but an "
                  + "important part of your delivery." );

    final JScrollPane infoScroll = new JScrollPane( info );
    infoScroll.setBorder( null );

    final JPanel top = new JPanel( new BorderLayout() );
    top.add( BorderLayout.NORTH,
             infoScroll );
    top.add( BorderLayout.CENTER,
             form );
    add( BorderLayout.NORTH,
         top );
    add( BorderLayout.CENTER,
         manifestAdditionsPane );
  }


  public class AdditionsModel
    implements
      RowEditor.Model,
      Iterable<Addition>
  {
    public AdditionsModel( final Project project )
    {
      super();
      this.project = project;
      for( final Addition item : project.getAdditions() )
        {
          this.additions.add( item );
        }
    }


    public void add( final Addition addition )
    {
      additions.add( addition );
    }


    @Override
    public int rowCount()
    {
      return additions.size();
    }


    @Override
    public int colCount()
    {
      return 1;
    }


    @Override
    public Component getComponent( final int row,
                                   final int col )
    {
      final Addition item = additions.get( row );
      Component c = components.get( item );
      if( c == null )
        {
          final FilePicker f = new FilePicker( item.getActualFile(),
                                               true,
                                               JFileChooser.FILES_ONLY,
                                               null,
                                               null );
          f.getJTextField().setToolTipText( "The actual file on disk to be stored" );

          final JTextField n = new JTextField( item.storeAs() );
          n.setToolTipText( "The name of the file in the JAR and/or Archive" );

          final JComboBox<Where> w = new JComboBox<>( Where.textArray() );
          w.setSelectedItem( item.where() );
          w.setToolTipText( "In which portion of your deliverable the file is to be stored" );

          final JPanel p = new JPanel( new BorderLayout() );
          final JPanel grid = new JPanel( new GridLayout( 1,
                                                          0 ) );
          components.put( item,
                          p );
          grid.add( f );
          grid.add( n );
          p.add( BorderLayout.CENTER,
                 grid );
          p.add( BorderLayout.EAST,
                 w );

          c = p;

          final ComponentListener unifiedListener = new ComponentListener()
          {
            @Override
            public void modified( final Object source )
            {
              if( source == f.getJTextField() )
                {
                  item.setActualFile( new File( f.getText() ) );
                  project.setAdditions( additions );
                }
              else if( source == n )
                {
                  item.setStoreAs( n.getText() );
                  project.setAdditions( additions );
                }
              else if( source == w )
                {
                  item.setWhere( (Where)w.getSelectedItem() );
                  project.setAdditions( additions );
                }
            }
          };
          f.getJTextField().addActionListener( unifiedListener );
          f.getJTextField().addFocusListener( unifiedListener );
          n.addActionListener( unifiedListener );
          n.addFocusListener( unifiedListener );
          w.addItemListener( unifiedListener );
        }
      return c;
    }


    @Override
    public void addRow()
    {
      additions.add( new Addition( new File( "" ),
                                   "",
                                   Where.BOTH ) );
      project.setAdditions( additions );
    }


    @Override
    public void delRow( final int rowIndex )
    {
      additions.remove( rowIndex );
      project.setAdditions( additions );
    }


    @Override
    public Iterator<Addition> iterator()
    {
      return additions.iterator();
    }

    private final Project project;
    private final List<Addition> additions = new ArrayList<>();
    private final Map<Addition,Component> components = new HashMap<>();
  }

  private static final long serialVersionUID = 4227648695959056400L;
}
