package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Insets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;

import com.ringlord.boxer2.Exclusion;
import com.ringlord.boxer2.Project;
import com.ringlord.swing.Form;


public class ConfigBinaries
  extends ConfigurationPanel
{
  public ConfigBinaries( final Project project )
  {
    super();

    final Form form = new Form( new Insets( 4,
                                            8,
                                            4,
                                            0 ) );

    form.setBorder( BorderFactory.createEmptyBorder( 8,
                                                     8,
                                                     8,
                                                     8 ) );
    form.setName( "ConfigBinaries" );

    // ==================================================

    final RowEditor exclusionEditor = new RowEditor( new ExclusionModel( project ),
                                                     "Add exclusion" );

    final JLabel exclusionLabel = new JLabel( "Exclude:",
                                              JLabel.TRAILING );
    exclusionLabel.setBorder( BorderFactory.createEmptyBorder( 7,
                                                               0,
                                                               0,
                                                               0 ) );
    exclusionLabel.setToolTipText( "Regular expression of filenames to exclude" );

    form.addRow( exclusionLabel,
                 true,
                 exclusionEditor );

    // ==================================================

    final JTextArea info = new JTextArea( 4,
                                          40 );
    info.setEditable( false );
    info.setLineWrap( true );
    info.setWrapStyleWord( true );
    info.setBackground( getBackground() );
    info.setText( "Binaries are your project's primary product, but some binaries that you MERGE from "
                  + "other projects and/or libraries may not be needed and only waste space in your "
                  + "main archive. Here is your chance (if you are careful) to trim away what you "
                  + "don't actually need to merge. Careful, though, as you can trigger runtime "
                  + "failures if you eliminate required dependencies!" );

    final JScrollPane infoScroll = new JScrollPane( info );
    infoScroll.setBorder( null );

    final JPanel top = new JPanel( new BorderLayout() );
    top.add( BorderLayout.NORTH,
             infoScroll );
    top.add( BorderLayout.CENTER,
             form );
    add( BorderLayout.NORTH,
         top );
  }


  public class ExclusionModel
    implements
      RowEditor.Model,
      Iterable<Exclusion>
  {
    public ExclusionModel( final Project project )
    {
      super();
      this.project = project;
      for( final Exclusion item : project.getBinariesExclusions() )
        {
          this.exclusions.add( item );
        }
    }


    public void add( final Exclusion exclusion )
    {
      exclusions.add( exclusion );
    }


    @Override
    public int rowCount()
    {
      return exclusions.size();
    }


    @Override
    public int colCount()
    {
      return 1;
    }


    @Override
    public Component getComponent( final int row,
                                   final int col )
    {
      final Exclusion item = exclusions.get( row );
      JTextField c = components.get( item );
      if( c == null )
        {
          final JTextField t = new JTextField( item.getPattern().toString() );
          final Color normalBackground = t.getBackground();
          t.setToolTipText( "<html>" + "Regular expression of file(s) to exclude.<br>" + "Example: /Test.*\\.class$" );
          components.put( item,
                          c = t );
          final ComponentListener unifiedListener = new ComponentListener()
          {
            @Override
            public void modified( final Object source )
            {
              try
                {
                  item.setPattern( Pattern.compile( t.getText() ) );
                  project.setBinariesExclusions( exclusions );
                  t.setBackground( normalBackground );
                  t.setToolTipText( null );
                }
              catch( final Exception x )
                {
                  t.setBackground( Color.yellow );
                  t.setToolTipText( "<html>" + "The regular expression is invalid:<br>" + x.getMessage() );
                }
            }
          };
          c.addActionListener( unifiedListener );
          c.addFocusListener( unifiedListener );
        }
      return c;
    }


    @Override
    public void addRow()
    {
      exclusions.add( new Exclusion() );
      project.setBinariesExclusions( exclusions );
    }


    @Override
    public void delRow( final int rowIndex )
    {
      exclusions.remove( rowIndex );
      project.setBinariesExclusions( exclusions );
    }


    @Override
    public Iterator<Exclusion> iterator()
    {
      return exclusions.iterator();
    }

    private final Project project;
    private final List<Exclusion> exclusions = new ArrayList<>();
    private final Map<Exclusion,JTextField> components = new HashMap<>();
  }

  private static final long serialVersionUID = 682314023883532799L;
}
