package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Insets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;

import com.ringlord.boxer2.Exclusion;
import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Project.Classpath;
import com.ringlord.swing.Form;


public class ConfigClasspath
  extends ConfigurationPanel
{
  public ConfigClasspath( final Project project )
  {
    super();

    final Form form = new Form( new Insets( 4,
                                            8,
                                            4,
                                            0 ) );

    form.setBorder( BorderFactory.createEmptyBorder( 8,
                                                     8,
                                                     8,
                                                     8 ) );
    form.setName( "ConfigLibraries" );

    // ==================================================

    final List<Classpath> classpaths = project.getClasspath();
    Collections.sort( classpaths );
    for( final Classpath classpath : classpaths )
      {
        createEntry( form,
                     project,
                     classpath );
      }

    // ==================================================

    final JTextArea info = new JTextArea( 3,
                                          40 );
    info.setEditable( false );
    info.setLineWrap( true );
    info.setWrapStyleWord( true );
    info.setBackground( getBackground() );
    info.setText( "The classpath refers to any collection of file resources in a directory, which are "
                  + "transferred to the archive in order to become available to your application. You "
                  + "may choose to exclude certain files from this set, but this is not recommended, "
                  + "as it increases the risk of runtime problems, even crashes." );

    final JScrollPane infoScroll = new JScrollPane( info );
    infoScroll.setBorder( null );

    final JPanel top = new JPanel( new BorderLayout() );
    top.add( BorderLayout.NORTH,
             infoScroll );
    top.add( BorderLayout.CENTER,
             form );
    add( BorderLayout.NORTH,
         top );
  }


  private void createEntry( final Form form,
                            final Project project,
                            final Classpath classpath )
  {
    final RowEditor exceptEditor = new RowEditor( new ExclusionModel( classpath ),
                                                  "Add exception" );

    final String pn = project.root().toString();
    final String ws = project.workspace().root().toString();
    String root = classpath.root().toString();
    if( root.startsWith( pn ) )
      {
        root = root.substring( pn.length() );
      }
    else if( root.startsWith( ws ) )
      {
        root = root.substring( ws.length() );
      }
    root = root + "/**";

    final JLabel rootLabel = new JLabel( root );
    rootLabel.setLabelFor( exceptEditor );
    rootLabel.setBorder( BorderFactory.createEmptyBorder( 8,
                                                          0,
                                                          0,
                                                          0 ) );
    final JLabel exceptLabel = new JLabel( "Except:",
                                           JLabel.TRAILING );
    exceptLabel.setBorder( BorderFactory.createEmptyBorder( 7,
                                                            0,
                                                            0,
                                                            0 ) );
    final Form exceptForm = new Form();
    exceptForm.addRow( exceptLabel,
                       true,
                       exceptEditor );

    form.addRow( rootLabel,
                 true,
                 exceptForm );
  }


  public class ExclusionModel
    implements
      RowEditor.Model,
      Iterable<Exclusion>
  {
    public ExclusionModel( final Classpath classpath )
    {
      super();
      this.classpath = classpath;
      for( final Exclusion item : classpath.getExclusions() )
        {
          this.exclusions.add( item );
        }
    }


    public void add( final Exclusion exclusion )
    {
      exclusions.add( exclusion );
    }


    @Override
    public int rowCount()
    {
      return exclusions.size();
    }


    @Override
    public int colCount()
    {
      return 1;
    }


    @Override
    public Component getComponent( final int row,
                                   final int col )
    {
      final Exclusion item = exclusions.get( row );
      JTextField c = components.get( item );
      if( c == null )
        {
          final JTextField t = new JTextField( item.getPattern().toString() );
          t.setToolTipText( "<html>" + "Regular expression of file(s) to exclude.<br>" + "Example: /Test.*\\.class$" );
          components.put( item,
                          c = t );
          final Color normalBackground = c.getBackground();
          final ComponentListener unifiedListener = new ComponentListener()
          {
            @Override
            public void modified( final Object source )
            {
              try
                {
                  item.setPattern( Pattern.compile( t.getText() ) );
                  classpath.setExclusions( exclusions );
                  t.setBackground( normalBackground );
                  t.setToolTipText( null );
                }
              catch( final Exception x )
                {
                  t.setBackground( Color.yellow );
                  t.setToolTipText( "<html>" + "The regular expression is invalid:<br>" + x.getMessage() );
                }
            }
          };
          c.addActionListener( unifiedListener );
          c.addFocusListener( unifiedListener );
        }
      return c;
    }


    @Override
    public void addRow()
    {
      exclusions.add( new Exclusion() );
      classpath.setExclusions( exclusions );
    }


    @Override
    public void delRow( final int rowIndex )
    {
      exclusions.remove( rowIndex );
      classpath.setExclusions( exclusions );
    }


    @Override
    public Iterator<Exclusion> iterator()
    {
      return exclusions.iterator();
    }

    private final Classpath classpath;
    private final List<Exclusion> exclusions = new ArrayList<>();
    private final Map<Exclusion,JTextField> components = new HashMap<>();
  }

  private static final long serialVersionUID = 8275807297511691099L;
}
