package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import com.ringlord.boxer2.Project;


public class ConfigJNLP
  extends ConfigurationPanel
{
  /**
   * <pre>
   *     <?xml version="1.0" encoding="UTF-8"?>
   * <jnlp spec="1.0+" codebase="http://ringlord.com/mindflayer/" href="mindflayer.jnlp">
   *   <information>
   *     <title>MindFlayer</title>
   *     <vendor>Ringlord Technologies</vendor>
   *     <description>Digital Art in Motion: It Flays Your Mind and Eats Your Brain</description>
   *     <description kind="short">Digital Art in Motion</description>
   *     <icon kind="splash" href="splash.png"/>
   *     <icon href="splash.png"/>
   *     <shortcut online="false">
   *       <desktop/>
   *       <menu submenu="Graphics"/>
   *     </shortcut>
   *     <offline-allowed/>
   *   </information>
   * <!--
   *   <security>
   *     <all-permissions/>
   *   </security>
   * -->
   *   <resources>
   *     <j2se version="1.5+" href="http://java.sun.com/products/autodl/j2se"/>
   *     <jar href="mindflayer.jar"/>
   *   </resources>
   *   <application-desc main-class="com.ringlord.mindflayer.Main">
   *     <argument>--proportional</argument>
   *   </application-desc>
   * </jnlp>
   * </pre>
   */
  public ConfigJNLP( final Project project )
  {
    super( new BorderLayout() );

    setBorder( BorderFactory.createEmptyBorder( 8,
                                                8,
                                                8,
                                                8 ) );
    setName( "ConfigJNLP" );

    // ==================================================

    this.header = new JLabel( " " );
    header.setBorder( BorderFactory.createEmptyBorder( 12,
                                                       8,
                                                       4,
                                                       0 ) );

    this.jnlpField = new JTextArea();
    jnlpField.setEditable( true );
    jnlpField.setLineWrap( false );
    jnlpField.setWrapStyleWord( false );
    jnlpField.setFont( new Font( "MonoSpaced",
                                 Font.PLAIN,
                                 jnlpField.getFont().getSize() ) );
    jnlpField.setText( project.getJNLP() );
    this.normalColor = jnlpField.getForeground();

    // ==================================================

    final ComponentListener combinedHandler = new ComponentListener()
    {
      @Override
      public void modified( final Object source )
      {
        if( source == jnlpField )
          {
            project.setJNLP( jnlpField.getText() );
          }
      }
    };

    jnlpField.addFocusListener( combinedHandler );

    final JPanel inputs = new JPanel( new BorderLayout() );
    inputs.add( BorderLayout.NORTH,
                header );
    inputs.add( BorderLayout.CENTER,
                new JScrollPane( jnlpField ) );

    // ==================================================

    final JTextArea info = new JTextArea( 4,
                                          40 );
    info.setEditable( false );
    info.setLineWrap( true );
    info.setWrapStyleWord( true );
    info.setBackground( getBackground() );
    info.setText( "A JNLP (Java Network Launch Protocol) file controls the launch parameters of a "
                  + "Java tool from a browser and subsequently from the desktop, by invoking the "
                  + ".jnlp file instead of the .jar. This requires the JAR to be digitally "
                  + "signed, which is a function controlled from the 'Main' tab, so enabling/"
                  + "disabling the creation of the JNLP file is also controlled on the 'Main' "
                  + "tab." );
    final JScrollPane infoScroll = new JScrollPane( info );
    infoScroll.setBorder( null );

    add( BorderLayout.NORTH,
         infoScroll );
    add( BorderLayout.CENTER,
         inputs );
  }


  public void setSelected( final boolean isSelected )
  {
    if( isSelected )
      {
        header.setText( "A JNLP file will be generated based on the settings below:" );
        jnlpField.setEditable( true );
        jnlpField.setForeground( normalColor );
      }
    else
      {
        header.setText( "A JNLP file will NOT be generated (enable it on the 'Main' tab)." );
        jnlpField.setEditable( false );
        jnlpField.setForeground( Color.gray );
      }
  }

  private final JLabel header;
  private final JTextArea jnlpField;
  private final Color normalColor;
  private static final long serialVersionUID = -5522538750912708989L;
}
