package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.FlowLayout;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.MutableComboBoxModel;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.filechooser.FileSystemView;

import com.ringlord.Filename;
import com.ringlord.Settings;
import com.ringlord.boxer2.Builder;
import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Project.Classpath;
import com.ringlord.boxer2.Project.ProjectReference;
import com.ringlord.boxer2.Util;
import com.ringlord.swing.ComboText;
import com.ringlord.swing.FilePicker;
import com.ringlord.swing.Form;
import com.ringlord.swing.Images;


public class ConfigMain
  extends ConfigurationPanel
{
  public ConfigMain( final Project project,
                     final ConfigJNLP configJNLP )
  {
    super( new BorderLayout() );

    final Form form = new Form( new Insets( 4,
                                            8,
                                            4,
                                            0 ) );
    form.setBorder( BorderFactory.createEmptyBorder( 8,
                                                     8,
                                                     8,
                                                     8 ) );
    form.setName( "ConfigMain" );

    // config is for global settings, not project-specific ones
    final Settings config = Util.getSettings();

    versionMaxLenShow = config.get( "version-preview-show-max",
                                    30 );
    versionMaxLenTrim = config.get( "version-preview-trim-max",
                                    Math.min( versionMaxLenShow,
                                              28 ) );

    // --------------------------------------------------
    // Description
    final JTextField descriptionField = new JTextField( project.getDescription() );
    final JLabel descriptionLabel = new JLabel( "Description:",
                                                JLabel.TRAILING );
    descriptionLabel.setLabelFor( descriptionField );
    setToolTipText( "Descriptive text or title for this project",
                    descriptionLabel,
                    descriptionField );

    // --------------------------------------------------
    // A sub-project is not buildable
    final JCheckBox isSubproject = new JCheckBox( "Depends on another project for packaging "
                                                      + "(not an independent project)",
                                                  project.isSubproject() );
    setToolTipText( "<html>"
                        + "Indicate that this project is<br>"
                        + "not intended to be packaged<br>"
                        + "by itself, but rather as part of<br>"
                        + "another project.",
                    isSubproject );

    // --------------------------------------------------
    // Working Directory (defaults to temporary directory)
    final File workingDirectory = new File( project.getWorkingDirectory() );
    final FilePicker workingDirectoryField = new FilePicker( workingDirectory,
                                                             true,
                                                             JFileChooser.DIRECTORIES_ONLY,
                                                             null,
                                                             null );
    final JLabel workingDirectoryLabel = new JLabel( "Working directory:",
                                                     JLabel.TRAILING );
    workingDirectoryLabel.setLabelFor( workingDirectoryField );
    setToolTipText( "<html>"
                        + "A directory where temporary files<br>"
                        + "can be extracted and assembled.<br>"
                        + "This requires as much as all the<br>"
                        + "space going into the deliverable.",
                    workingDirectoryLabel,
                    workingDirectoryField );

    // --------------------------------------------------
    // Output JAR file
    final File jarFile = new File( project.getOutputJarFile() );
    final FilePicker jarFilenameField = new FilePicker( jarFile,
                                                        true,
                                                        JFileChooser.FILES_ONLY,
                                                        null,
                                                        null );
    final JLabel jarFilenameLabel = new JLabel( "JAR filename:",
                                                JLabel.TRAILING );
    jarFilenameLabel.setLabelFor( jarFilenameField );
    setToolTipText( "<html>"
                        + "The executable JAR file to create.<br>"
                        + "~/ represents your home directory.<br>"
                        + "${VERSION} is the version tag.<br>"
                        + "${USER} is your login name.<br>"
                        + "${OS} is the operating system name<br>"
                        + "${BUILD} is the current build number.",
                    jarFilenameLabel,
                    jarFilenameField );

    // --------------------------------------------------
    // Build JNLP
    final JCheckBox buildJNLP = new JCheckBox( "Build JNLP (Java Network Launch Protocol) file, "
                                                   + "see JNLP tab above",
                                               project.getBuildJNLP() );
    setToolTipText( "<html>"
                        + "Launching Java tools through a URL or from the<br>"
                        + "desktop requires the JAR to be properly signed,<br>"
                        + "and the .jnlp file contents to be configured on<br>"
                        + "the 'JNLP' tab above.",
                    buildJNLP );
    configJNLP.setSelected( buildJNLP.isSelected() );

    // --------------------------------------------------
    // Sign the JAR?
    final JLabel missingJarsignerLabel = new JLabel( "\u2014 No JarSigner, click here to select the program" );
    missingJarsignerLabel.setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
    missingJarsignerLabel.setBorder( BorderFactory.createEmptyBorder( 0,
                                                                      12,
                                                                      0,
                                                                      0 ) );
    missingJarsignerLabel.setForeground( Color.blue );
    final JCheckBox signJar = new JCheckBox( "Sign JAR",
                                             project.getSignJAR() || project.getBuildJNLP() );
    if( project.getBuildJNLP() )
      {
        signJar.setEnabled( false );
      }
    setToolTipText( "<html>"
                        + "Digitally sign the JAR file contents with your private key,<br>"
                        + "which provides assurance to recipients of the JAR file's<br>"
                        + "origin. This is required for JNLP-delivered applications.",
                    signJar );
    final JPanel signJarPane = new JPanel( new BorderLayout() );
    signJarPane.add( BorderLayout.WEST,
                     signJar );
    signJarPane.add( BorderLayout.CENTER,
                     missingJarsignerLabel );

    // --------------------------------------------------
    // File containing keystore secrets
    final File keystoreSecretsFile = new File( project.getSigningPropertiesFile() );
    final FilePicker keystoreSecretsField = new FilePicker( keystoreSecretsFile,
                                                            true,
                                                            JFileChooser.FILES_ONLY,
                                                            null,
                                                            null );
    final JLabel keystoreSecretsLabel = new JLabel( "Keystore secrets file:",
                                                    JLabel.TRAILING );
    keystoreSecretsLabel.setLabelFor( keystoreSecretsField );
    setToolTipText( "File containing secrets for accessing the keystore",
                    keystoreSecretsLabel,
                    keystoreSecretsField );
    final JButton keystoreSecretsEdit = new JButton( "Edit\u2026",
                                                     Images.getIcon( "/img/pencil.png" ) );
    setToolTipText( "Edit the secrets needed to access the keystore",
                    keystoreSecretsEdit );

    final JTextField tsaField = new JTextField( 32 );
    final JLabel tsaLabel = new JLabel( "Timestamp Authority URL:",
                                        JLabel.TRAILING );
    tsaLabel.setLabelFor( tsaField );
    setToolTipText( "<html>"
                        + "Applying a timestamp when you sign a JAR allows<br>"
                        + "you to prove that you signed the JAR while your<br>"
                        + "code signing certificate was still valid. Your JAR<br>"
                        + "can be validated after the certificate has expired,<br>"
                        + "which prolongs the life of your software. Search<br>"
                        + "'<i>timestamp server URL</i>' for a TSA Server that you<br>"
                        + "can use (enter its URL here). Leave blank to skip.",
                    tsaLabel,
                    tsaField );

    final Form keystoreForm = new Form();
    keystoreForm.setName( "ConfigMain.keystoreSecretsForm" );
    keystoreForm.addRow( keystoreSecretsLabel,
                         keystoreSecretsField,
                         keystoreSecretsEdit );
    keystoreForm.addRow( tsaLabel,
                         tsaField,
                         2,
                         null );

    // --------------------------------------------------
    // Main class

    final JComboBox<String> mainClassField = new JComboBox<>();
    mainClassField.setEditable( true );
    final String chosenEntryPoint = project.getEntryPointName();
    if( chosenEntryPoint != null )
      {
        // Make this the first item
        mainClassField.addItem( chosenEntryPoint );
      }
    int uppermostLevel = Integer.MAX_VALUE;
    String uppermostClass = null;
    final List<Class<?>> javaEntryPoints = project.getEntryPoints();
    for( final Class<?> entryPoint : javaEntryPoints )
      {
        final String entryPointName = entryPoint.getName();
        if( (chosenEntryPoint == null) || !chosenEntryPoint.equals( entryPointName ) )
          {
            final String name = entryPoint.getName();
            mainClassField.addItem( name );

            if( chosenEntryPoint == null )
              {
                int dots = 0;
                final int len = name.length();
                for( int i = 0; i < len; i++ )
                  {
                    if( name.charAt( i ) == '.' )
                      {
                        dots++;
                      }
                  }
                if( dots < uppermostLevel )
                  {
                    uppermostLevel = dots;
                    uppermostClass = name;
                  }
              }
          }
      }
    if( (chosenEntryPoint == null) && (uppermostClass != null) )
      {
        // Pick the highest class in the hierarchy
        mainClassField.setSelectedItem( uppermostClass );
        project.setEntryPointName( uppermostClass );
      }
    final JLabel mainClassLabel = new JLabel( "Entry point:",
                                              JLabel.TRAILING );
    mainClassLabel.setLabelFor( mainClassField );
    setToolTipText( "<html>"
                        + "The main class, where execution begins. The available<br>"
                        + "choices include all classes with a method signature<br>"
                        + "<br>"
                        + "&nbsp;&nbsp;&nbsp;&nbsp;<code>public static void main( String[] )</code><br>"
                        + "<br>"
                        + "You may enter a different classname; the value in this<br>"
                        + "field becomes the JAR manifest's 'Main-Class' parameter.",
                    mainClassLabel,
                    mainClassField );

    // --------------------------------------------------

    final JTextField splashImageField = new JTextField( project.getSplashImageName() );
    final JLabel splashImageLabel = new JLabel( "Splash image name:",
                                                JLabel.TRAILING );
    splashImageLabel.setLabelFor( splashImageField );
    setToolTipText( "<html>"
                        + "The Java SplashScreen image filename as<br>"
                        + "packaged in the JAR file. A leading '/' will<br>"
                        + "be supplied if you omit it.",
                    splashImageLabel,
                    splashImageField );
    final JButton pickSplashImage = new JButton( "Pick\u2026",
                                                 Images.getIcon( "/img/folder.png" ) );
    pickSplashImage.setToolTipText( "Pick file" );
    final JPanel splashImagePane = new JPanel( new BorderLayout() );
    splashImagePane.add( BorderLayout.CENTER,
                         splashImageField );
    splashImagePane.add( BorderLayout.EAST,
                         pickSplashImage );

    // --------------------------------------------------

    final JCheckBox mkDistro = new JCheckBox( "Create distribution archive file",
                                              project.createDistributionFile() );
    setToolTipText( "<html>"
                        + "Generate a distributable archive file<br>"
                        + "containing the JAR and all dependent<br>"
                        + "JARs (if your project has externally<br>"
                        + "referenced library JARs).",
                    mkDistro );

    // --------------------------------------------------
    // Output archive file
    final File distroFile = new File( project.getDistributionFile() );
    final FilePicker distroField = new FilePicker( distroFile,
                                                   true,
                                                   JFileChooser.FILES_ONLY,
                                                   null,
                                                   null );
    final JLabel distroLabel = new JLabel( "Archive filename:",
                                           JLabel.TRAILING );
    distroLabel.setLabelFor( distroField );
    setToolTipText( "<html>"
                        + "The distributable archive file to create.<br>"
                        + "~/ represents your home directory.<br>"
                        + "${VERSION} is the version tag.<br>"
                        + "${USER} is your login name.<br>"
                        + "${OS} is the operating system name<br>"
                        + "${BUILD} is the current build number.",
                    distroLabel,
                    distroField );

    final JTextField distroSubdirField = new JTextField( project.getDistributionSubdir() );
    final JLabel distroSubdirLabel = new JLabel( "In-archive root:",
                                                 JLabel.TRAILING );
    distroSubdirLabel.setLabelFor( distroSubdirField );
    setToolTipText( "<html>"
                        + "Subdirectory within distribution archive. It<br>"
                        + "defaults to the project name and resolves<br>"
                        + "${VERSION}, ${USER}, ${OS}, and ${BUILD}.",
                    distroSubdirLabel,
                    distroSubdirField );

    final Form distroForm = new Form();
    distroForm.setName( "ConfigMain.distroForm" );
    distroForm.addRow( distroLabel,
                       distroField );
    distroForm.addRow( distroSubdirLabel,
                       distroSubdirField );

    // --------------------------------------------------
    // Obtain the names of MessageDigest algorithms. Eliminate unusual
    // names like "MD5 InstalledIn" or generic "SHA" without specific
    // bit size included.
    final List<String> digestNames = Util.getMessageDigestAlgorithms( project );
    Collections.sort( digestNames );
    final JPanel digestsField = new JPanel( new FlowLayout( FlowLayout.LEFT,
                                                            0,
                                                            0 ) );
    final List<JCheckBox> digestChoices = new ArrayList<>();
    boolean isFirst = true;
    for( final String algorithm : digestNames )
      {
        final String ext = "." + algorithm.replace( "-",
                                                    "" ).toLowerCase();
        final JCheckBox b = new JCheckBox( ext,
                                           project.isDigestPicked( algorithm ) );
        if( Util.isLegacyMessageDigestAlgorithm( algorithm ) )
          {
            b.setToolTipText( "<html>" +
                              "This appears to be a legacy algorithm, and<br>" +
                              "may become unavailable in the future.<br>" +
                              "<br>" +
                              "Create a digest signature file with " +
                              ext +
                              "<br>" +
                              "filename extension, based on the " +
                              algorithm +
                              "<br>" +
                              "algorithm." );
            b.setForeground( Color.red.darker() );
          }
        else
          {
            b.setToolTipText( "<html>" +
                              "Create a digest signature file<br>" +
                              "with " +
                              ext +
                              " filename extension,<br>" +
                              "based on the " +
                              algorithm +
                              " algorithm." );
          }
        b.setName( algorithm );
        digestChoices.add( b );
        b.setName( algorithm );
        if( isFirst )
          {
            isFirst = false;
          }
        else
          {
            digestsField.add( new JLabel( "     " ) ); // 5 spaces
          }
        digestsField.add( b );
      }
    final JLabel digestsLabel = new JLabel( "Digest signature files:" );
    digestsLabel.setLabelFor( digestsField );
    digestsLabel.setToolTipText( "<html>"
                                 + "Create digest signature file(s) to be<br>"
                                 + "uploaded along with the JAR to allow<br>"
                                 + "verification of the JAR's integrity." );

    // --------------------------------------------------
    // ${VERSION} tag
    final JLabel versionExtractLabel = new JLabel( "Define ${VERSION} tag:",
                                                   JLabel.TRAILING );
    versionExtractLabel.setBorder( BorderFactory.createEmptyBorder( 7,
                                                                    0,
                                                                    0,
                                                                    0 ) );
    versionExtractLabel.setToolTipText( "<html>"
                                        + "An optional ${VERSION} tag can be<br>"
                                        + "used in JAR and Archive filenames." );

    final File versionSource = new File( project.getVersionExtractFilename() );
    final FilePicker versionFilenameField = new FilePicker( versionSource,
                                                            true,
                                                            JFileChooser.FILES_ONLY,
                                                            null,
                                                            null );
    final JLabel versionFilenameLabel = new JLabel( "From source file:",
                                                    JLabel.TRAILING );
    versionFilenameLabel.setLabelFor( versionFilenameField );
    setToolTipText( "<html>"
                        + "The source file from which information is<br>"
                        + "to be extracted that forms the value of a<br>"
                        + "${VERSION} tag that can be used in the<br>"
                        + "JAR and Archive filenames. A leading '/'<br>"
                        + "refers to the workspace, otherwise the<br>"
                        + "path is relative to the project.",
                    versionFilenameLabel,
                    versionFilenameField );

    final MutableComboBoxModel<String> versionRegexModel = new DefaultComboBoxModel<>();
    final ComboText versionRegexField = new ComboText( versionRegexModel,
                                                       20 );
    versionRegexField.setEditable( true );
    versionRegexField.add( project.getVersionExtractRegexes() );
    final JLabel versionRegexLabel = new JLabel( "Extract by Regex:",
                                                 JLabel.TRAILING );
    versionRegexLabel.setLabelFor( versionRegexField );
    setToolTipText( "<html>"
                        + "The regular expression that finds one or more<br>"
                        + "parts of the ${VERSION} tag in the source, and<br>"
                        + "each part must be enclosed in parentheses to<br>"
                        + "mark it as a 'group' (see Assembly for more)",
                    versionRegexLabel,
                    versionRegexField );

    final MutableComboBoxModel<String> versionAssemblyModel = new DefaultComboBoxModel<>();
    final ComboText versionAssemblyField = new ComboText( versionAssemblyModel,
                                                          20 );
    versionAssemblyField.setEditable( true );
    versionAssemblyField.add( project.getVersionExtractAssemblies() );
    final JLabel versionAssemblyLabel = new JLabel( "Assemble tag:",
                                                    JLabel.TRAILING );
    versionAssemblyLabel.setLabelFor( versionAssemblyField );
    setToolTipText( "<html>"
                        + "The text to become the ${VERSION}.<br>"
                        + "References to the 'groups' from the<br>"
                        + "regex takes the form $(1), $(2), etc.<br>"
                        + "You need not reference all groups.",
                    versionAssemblyLabel,
                    versionAssemblyField );

    final JLabel versionPreview = new JLabel( "" );
    versionPreview.setToolTipText( "<html>" + "Preview the value of<br>the ${VERSION} tag" );

    final JPanel versionPreviewPanel = new JPanel( new BorderLayout() );
    versionPreviewPanel.add( BorderLayout.CENTER,
                             versionAssemblyField );
    versionPreviewPanel.add( BorderLayout.EAST,
                             versionPreview );

    final Form versionExtractForm = new Form();
    versionExtractForm.setName( "ConfigMain.versionExtract" );

    versionExtractForm.addRow( versionFilenameLabel,
                               versionFilenameField );
    versionExtractForm.addRow( versionRegexLabel,
                               versionRegexField );
    versionExtractForm.addRow( versionAssemblyLabel,
                               versionPreviewPanel );

    // ==================================================

    form.addRow( descriptionLabel,
                 descriptionField );

    form.addRow( null,
                 isSubproject );

    form.addRow( jarFilenameLabel,
                 jarFilenameField );

    form.addRow( null,
                 buildJNLP );

    form.addRow( null,
                 signJarPane );

    form.addRow( null,
                 keystoreForm );

    form.addRow( mainClassLabel,
                 mainClassField );

    form.addRow( splashImageLabel,
                 splashImagePane );

    form.addRow( null,
                 mkDistro );

    form.addRow( null,
                 distroForm );

    form.addRow( digestsLabel,
                 digestsField );

    form.addRow( workingDirectoryLabel,
                 workingDirectoryField );

    form.addRow( versionExtractLabel,
                 true,
                 versionExtractForm );

    // ==================================================

    final ComponentListener combinedHandler = new ComponentListener()
    {
      @Override
      public void modified( final Object source )
      {
        if( source == descriptionField )
          {
            project.setDescription( descriptionField.getText() );
          }
        else if( source == isSubproject )
          {
            project.setSubproject( isSubproject.isSelected() );
          }
        else if( source == jarFilenameField.getJTextField() )
          {
            String name = jarFilenameField.getText().trim();
            if( name.length() == 0 )
              {
                name = new File( project.root(),
                                 project.name() + ".jar" ).toString();
              }
            project.setOutputJarName( name );
          }
        else if( source == keystoreSecretsField.getJTextField() )
          {
            final String name = keystoreSecretsField.getText().trim();
            project.setSigningPropertiesFile( name.length() == 0
                ? null
                : name );
          }
        else if( source == buildJNLP )
          {
            final boolean isSelected = buildJNLP.isSelected();
            project.setBuildJNLP( isSelected );
            if( isSelected )
              {
                signJar.setEnabled( false );
                signJar.setSelected( true );
                configJNLP.setSelected( true );
              }
            else
              {
                configJNLP.setSelected( false );
                signJar.setEnabled( true );
              }
          }
        else if( source == signJar )
          {
            final boolean isSelected = signJar.isSelected();
            project.setSignJAR( isSelected );
            keystoreForm.setVisible( isSelected );
            form.revalidate(); // important
            if( signJar.isEnabled() )
              {
                signJar.requestFocusInWindow();
              }
          }
        else if( source == mkDistro )
          {
            final boolean isSelected = mkDistro.isSelected();
            project.setCreateDistributionFile( isSelected );
            distroForm.setVisible( isSelected );
            form.revalidate(); // important
            if( mkDistro.isEnabled() )
              {
                mkDistro.requestFocusInWindow();
              }
          }
        else if( source == distroField.getJTextField() )
          {
            String name = distroField.getText().trim();
            if( name.length() == 0 )
              {
                name = new File( project.getWorkingDirectory(),
                                 project.name() + ".tar.gz" ).toString();
              }
            project.setDistributionName( name );
          }
        else if( source == distroSubdirField )
          {
            final String name = distroSubdirField.getText().trim();
            project.setDistributionSubdir( name );
          }
        else if( digestChoices.contains( source ) )
          {
            final JCheckBox d = (JCheckBox)source;
            final boolean isSelected = d.isSelected();
            project.setDigestPicked( d.getName(),
                                     isSelected );
          }
        else if( source == mainClassField )
          {
            project.setEntryPointName( mainClassField.getSelectedItem().toString() );
          }
        else if( source == splashImageField )
          {
            boolean updateField = false;
            String fname = splashImageField.getText();
            if( (fname.length() > 0) && !fname.startsWith( "/" ) )
              {
                fname = "/" + fname;
                updateField = true;
              }
            project.setSplashImageName( fname.length() == 0
                ? null
                : fname );
            if( updateField )
              {
                splashImageField.setText( fname );
              }
          }
        else if( source == pickSplashImage )
          {
            if( splashImageChooser == null )
              {
                final File home = getProjectClasspaths( project,
                                                        rootsList );
                final File[] roots = new File[rootsList.size()];
                rootsList.toArray( roots );
                final FileSystemView fsv = new ProjectClasspathView( roots,
                                                                     home );
                splashImageChooser = new JFileChooser( fsv );
                splashImageChooser.setDialogTitle( "Pick Splash Image File" );
                splashImageChooser.setFileSelectionMode( JFileChooser.FILES_ONLY );
              }
            if( splashImageChooser.showOpenDialog( ConfigMain.this ) == JFileChooser.APPROVE_OPTION )
              {
                final File splashImageFile = splashImageChooser.getSelectedFile();
                String fname = splashImageFile.toString();
                for( final File root : rootsList )
                  {
                    if( fname.startsWith( root.toString() ) )
                      {
                        fname = fname.substring( root.toString().length() );
                        break;
                      }
                  }
                splashImageField.setText( fname );
                project.setSplashImageName( fname );
              }
          }
      }


      private File getProjectClasspaths( final Project project,
                                         final List<File> rootsList )
      {
        final Set<Project> done = new HashSet<>();
        getProjectClasspaths( project,
                              rootsList,
                              done );
        return (rootsList.isEmpty()
            ? null
            : rootsList.get( 0 ));
      }


      private void getProjectClasspaths( final Project project,
                                         final List<File> rootsList,
                                         final Set<Project> done )
      {
        done.add( project );
        final List<Classpath> classpaths = project.getClasspath();
        if( classpaths != null )
          {
            for( final Classpath cp : classpaths )
              {
                rootsList.add( cp.root() );
              }
          }
        done.add( project );
        final List<ProjectReference> projectRefs = project.getProjects();
        if( projectRefs != null )
          {
            for( final ProjectReference pRef : projectRefs )
              {
                final Project p = pRef.project();
                if( !done.contains( p ) )
                  {
                    getProjectClasspaths( p,
                                          rootsList,
                                          done );
                  }
              }
          }
      }

      private JFileChooser splashImageChooser;
      final List<File> rootsList = new ArrayList<>();
    };

    descriptionField.addFocusListener( combinedHandler );
    descriptionField.addActionListener( combinedHandler );
    isSubproject.addItemListener( combinedHandler );
    jarFilenameField.getJTextField().addActionListener( combinedHandler );
    jarFilenameField.getJTextField().addFocusListener( combinedHandler );
    keystoreSecretsField.getJTextField().addActionListener( combinedHandler );
    keystoreSecretsField.getJTextField().addFocusListener( combinedHandler );
    buildJNLP.addItemListener( combinedHandler );
    signJar.addItemListener( combinedHandler );
    splashImageField.addActionListener( combinedHandler );
    splashImageField.addFocusListener( combinedHandler );
    pickSplashImage.addActionListener( combinedHandler );
    mkDistro.addItemListener( combinedHandler );
    distroField.getJTextField().addActionListener( combinedHandler );
    distroField.getJTextField().addFocusListener( combinedHandler );
    distroSubdirField.addActionListener( combinedHandler );
    distroSubdirField.addFocusListener( combinedHandler );
    mainClassField.addActionListener( combinedHandler );
    for( final JCheckBox d : digestChoices )
      {
        d.addItemListener( combinedHandler );
      }

    // --------------------------------------------------

    keystoreSecretsEdit.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        try
          {
            final JDialog dialog = new JDialog( (JFrame)null,
                                                true );
            final File curFile = Filename.resolveTilde( keystoreSecretsField.getText() );
            final SigningPropertiesEditor editor = new SigningPropertiesEditor( curFile,
                                                                                dialog );
            final List<Image> imageIcons = new ArrayList<>();
            imageIcons.add( Images.getImage( "/img/boxer2-96.png" ) );
            imageIcons.add( Images.getImage( "/img/boxer2-64.png" ) );
            imageIcons.add( Images.getImage( "/img/boxer2-48.png" ) );
            imageIcons.add( Images.getImage( "/img/boxer2-32.png" ) );
            imageIcons.add( Images.getImage( "/img/boxer2-24.png" ) );
            imageIcons.add( Images.getImage( "/img/boxer2-16.png" ) );
            dialog.setIconImages( imageIcons );
            dialog.setTitle( "Editing " + curFile );
            dialog.add( new JScrollPane( editor ) );
            dialog.pack();
            dialog.setLocationRelativeTo( keystoreSecretsField );
            dialog.setVisible( true );
            dialog.dispose();
          }
        catch( final IllegalStateException x )
          {
            // The file contains encrypted contents but the user failed to
            // provide the correct password, and selected to cancel the
            // operation.
          }
      }
    } );

    // --------------------------------------------------

    final DocumentListener versionUpdater = new DocumentListener()
    {
      @Override
      public void changedUpdate( final DocumentEvent e )
      {
      }


      @Override
      public void insertUpdate( final DocumentEvent e )
      {
        updatePreview();
      }


      @Override
      public void removeUpdate( final DocumentEvent e )
      {
        updatePreview();
      }


      private void updatePreview()
      {
        versionPreview.setText( updateVersionPreview( project,
                                                      versionFilenameField.getText(),
                                                      versionRegexField.getText(),
                                                      versionAssemblyField.getText() ) );
      }
    };

    versionFilenameField.getJTextField().getDocument().addDocumentListener( versionUpdater );
    versionRegexField.addDocumentListener( versionUpdater );
    versionAssemblyField.addDocumentListener( versionUpdater );

    versionFilenameField.getJTextField().addFocusListener( new FocusListener()
    {
      @Override
      public void focusGained( final FocusEvent arg0 )
      {
      }


      @Override
      public void focusLost( final FocusEvent arg0 )
      {
        project.setVersionExtractFilename( versionFilenameField.getText() );
      }
    } );

    versionRegexField.addFocusListener( new FocusListener()
    {
      @Override
      public void focusGained( final FocusEvent e )
      {
      }


      @Override
      public void focusLost( final FocusEvent e )
      {
        final List<String> versionRegexes = new ArrayList<>();
        final int max = Math.min( config.get( "version-remember-regex-max",
                                              20 ),
                                  versionRegexModel.getSize() );
        for( int i = 0; i < max; i++ )
          {
            versionRegexes.add( versionRegexModel.getElementAt( i ) );
          }
        project.setVersionExtractRegexes( versionRegexes );
      }
    } );

    versionAssemblyField.addFocusListener( new FocusListener()
    {
      @Override
      public void focusGained( final FocusEvent e )
      {
      }


      @Override
      public void focusLost( final FocusEvent e )
      {
        final List<String> versionAssemblies = new ArrayList<>();
        final int max = Math.min( config.get( "version-remember-assembly-max",
                                              20 ),
                                  versionAssemblyModel.getSize() );
        for( int i = 0; i < max; i++ )
          {
            versionAssemblies.add( versionAssemblyModel.getElementAt( i ) );
          }
        project.setVersionExtractAssemblies( versionAssemblies );
      }
    } );

    // ==================================================

    versionPreview.setText( updateVersionPreview( project,
                                                  versionFilenameField.getText(),
                                                  versionRegexField.getText(),
                                                  versionAssemblyField.getText() ) );

    keystoreForm.setVisible( signJar.isSelected() );
    distroForm.setVisible( mkDistro.isSelected() );
    form.revalidate(); // important

    if( Builder.getJarSignerBinary() == null )
      {
        missingJarsignerLabel.addMouseListener( new MouseAdapter()
        {
          @Override
          public void mouseClicked( final MouseEvent e )
          {
            if( fileChooser == null )
              {
                fileChooser = new JFileChooser();
                fileChooser.setCurrentDirectory( new File( System.getProperty( "user.home" ) ) );
              }
            fileChooser.setDialogTitle( "Pick JarSigner Program" );
            if( fileChooser.showOpenDialog( ConfigMain.this ) == JFileChooser.APPROVE_OPTION )
              {
                final String binary = fileChooser.getSelectedFile().toString();
                try
                  {
                    if( Builder.checkJarsigner( binary ) )
                      {
                        missingJarsignerLabel.setVisible( false );
                        missingJarsignerLabel.removeMouseListener( this );
                        Util.getSettings().put( "jarsigner-binary",
                                                binary );
                        return;
                      }
                    // fall throough
                  }
                catch( final IOException x )
                  {
                    // fall through
                  }
                JOptionPane.showMessageDialog( ConfigMain.this,
                                               "The selected command does not\n" + "support the needed parameters!",
                                               "Bad JarSigner",
                                               JOptionPane.ERROR_MESSAGE );
              }
          }

          private JFileChooser fileChooser;
        } );
      }
    else
      {
        missingJarsignerLabel.setVisible( false );
      }

    /**
     * <pre>
     * working directory
     *
     * output .jar file
     * [x] sign .jar
     *     secrets file [___________] [edit] [pick]
     *     [x] Time Stamp Authority URL [_________]
     * [x] signature (dynamic)
     *
     * [x] build distribution archive
     *     filename [_____________][.tar.gz|v] [dir]
     *     subdir in archive [(project)___________]
     *
     * heading [____________] (probably should be near top
     * descript [_______________] near top?
     *
     * ${VERSION} source file [____________________] [pick]
     *            extract regex [_________________]
     *            transform [___________] <<sample>>
     * </pre>
     */

    add( BorderLayout.NORTH,
         form );
  }


  private static String updateVersionPreview( final Project project,
                                              final String filename,
                                              final String regex,
                                              final String assembly )
  {
    final String version = Util.getVersion( project,
                                            filename,
                                            regex,
                                            assembly );
    // \u21d2 is an arrow such as =>
    return version == null
        ? ""
        : (" \u21d2 " + (version.length() > versionMaxLenShow
            ? (version.substring( 0,
                                  versionMaxLenTrim ) + "\u2026")
            : version));
  }

  private static int versionMaxLenShow = 30;
  private static int versionMaxLenTrim = 28;
  private static final long serialVersionUID = -8057937483182427756L;
}
