package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Insets;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import javax.swing.BorderFactory;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;

import com.ringlord.boxer2.Exclusion;
import com.ringlord.boxer2.FileHandlingOption;
import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Project.Sources;
import com.ringlord.swing.Form;


public class ConfigSources
  extends ConfigurationPanel
{
  public ConfigSources( final Project project )
  {
    super();

    final Form form = new Form( new Insets( 4,
                                            8,
                                            4,
                                            0 ) );

    form.setBorder( BorderFactory.createEmptyBorder( 8,
                                                     8,
                                                     8,
                                                     8 ) );
    form.setName( "ConfigSources" );

    // ==================================================

    final JTextField sourceRootField = new JTextField( project.getSourceRoot() );
    final JLabel sourceRootLabel = new JLabel( "Source path:",
                                               JLabel.TRAILING );
    sourceRootLabel.setLabelFor( sourceRootField );
    setToolTipText( "<html>"
                        + "The directory under which sources are stored in<br>"
                        + "the main archive (if you choose to include your<br>"
                        + "project's source code as indicated below).<br>"
                        + "<br>"
                        + "Source code compresses well, so it won't grow<br>"
                        + "the archive by much, but ensures that FLOSS<br>"
                        + "projects always carry their source!",
                    sourceRootLabel,
                    sourceRootField );
    final ComponentListener unifiedListener = new ComponentListener()
    {
      @Override
      public void modified( final Object source )
      {
        project.setSourceRoot( sourceRootField.getText().trim() );
      }
    };
    sourceRootField.addActionListener( unifiedListener );
    sourceRootField.addFocusListener( unifiedListener );

    form.addRow( sourceRootLabel,
                 sourceRootField );

    final List<Sources> projectSources = project.getSources();
    for( final Sources sources : projectSources )
      {
        final Form filter = new Form();
        filter.setName( "ConfigSources." + sources.root() );

        final JComboBox<FileHandlingOption> handling = new JComboBox<>( FileHandlingOption.textArray() );
        handling.setRenderer( renderer );
        handling.setSelectedItem( sources.getHandling() );
        handling.setToolTipText( "<html>"
                                 + "Select INCLUDE to include source code from<br>"
                                 + "your project and sub-projects that also allow<br>"
                                 + "to include sources. If you specify IGNORE,<br>"
                                 + "then NOT source code will be included, even<br>"
                                 + "if sub-projects allow source to be included." );

        final RowEditor exclusionEditor = new RowEditor( new ExclusionModel( sources ),
                                                         "Add exception" );

        final JLabel exclusionLabel = new JLabel( "Except:",
                                                  JLabel.TRAILING );
        exclusionLabel.setBorder( BorderFactory.createEmptyBorder( 7,
                                                                   0,
                                                                   0,
                                                                   0 ) );
        exclusionLabel.setToolTipText( "<html>"
                                       + "Regular expression of filenames to exclude. Examples:<br>"
                                       + "<code>org/example/Demo.java</code><br>"
                                       + "<code>.*/Test.*\\.java</code>" );
        filter.addRow( exclusionLabel,
                       true,
                       exclusionEditor );

        final String ws = project.root().toString();
        String root = sources.root().toString();
        if( root.startsWith( ws ) )
          {
            root = root.substring( ws.length() );
            if( root.startsWith( "/" ) )
              {
                root = root.substring( 1 );
              }
          }
        root += "/**";

        if( handling.getSelectedItem() == FileHandlingOption.IGNORE )
          {
            exclusionLabel.setVisible( false );
            exclusionEditor.setVisible( false );
          }
        handling.addItemListener( new ItemListener()
        {
          @Override
          public void itemStateChanged( final ItemEvent e )
          {
            final Object selection = handling.getSelectedItem();
            if( selection == FileHandlingOption.IGNORE )
              {
                exclusionLabel.setVisible( false );
                exclusionEditor.setVisible( false );
                sources.setHandling( FileHandlingOption.IGNORE );
              }
            else if( selection == FileHandlingOption.INCLUDE )
              {
                exclusionLabel.setVisible( true );
                exclusionEditor.setVisible( true );
                sources.setHandling( FileHandlingOption.INCLUDE );
              }
            else
              {
                // TODO log this issue
                System.err.println( "Unhandled FileHandlingOption: " + selection );
              }
            filter.revalidate();
          }
        } );

        form.addRow( new JLabel( root,
                                 JLabel.TRAILING ),
                     handling );
        form.addRow( null,
                     filter );
      }

    // ==================================================

    final JTextArea info = new JTextArea( 4,
                                          40 );
    info.setEditable( false );
    info.setLineWrap( true );
    info.setWrapStyleWord( true );
    info.setBackground( getBackground() );
    info.setText( "If you are building Free/Libre/Open Source Software (FLOSS) tnen you may want "
                  + "to select below to INCLUDE the source code in the JAR, which ensures maximum "
                  + "convenience to the recipient (who can treat the JAR like a ZIP file in order "
                  + "to extract file files). If you do not want to include source code in the JAR "
                  + "(perhaps you are distributing it otherwise, or you are developing proprietary "
                  + "software), then ensure that you select IGNORE below." );

    final JScrollPane infoScroll = new JScrollPane( info );
    infoScroll.setBorder( null );

    final JPanel top = new JPanel( new BorderLayout() );
    top.add( BorderLayout.NORTH,
             infoScroll );
    top.add( BorderLayout.CENTER,
             form );
    add( BorderLayout.NORTH,
         top );
  }


  public class ExclusionModel
    implements
      RowEditor.Model,
      Iterable<Exclusion>
  {
    public ExclusionModel( final Sources sources )
    {
      super();
      this.sources = sources;
      for( final Exclusion item : sources.getExclusions() )
        {
          this.exclusions.add( item );
        }
    }


    public void add( final Exclusion exclusion )
    {
      exclusions.add( exclusion );
    }


    @Override
    public int rowCount()
    {
      return exclusions.size();
    }


    @Override
    public int colCount()
    {
      return 1;
    }


    @Override
    public Component getComponent( final int row,
                                   final int col )
    {
      final Exclusion item = exclusions.get( row );
      JTextField c = components.get( item );
      if( c == null )
        {
          final JTextField t = new JTextField( item.getPattern().toString() );
          t.setToolTipText( "<html>" + "Regular expression of file(s) to exclude.<br>" + "Example: /Test.*\\.class$" );
          components.put( item,
                          c = t );
          final Color normalBackground = c.getBackground();
          final ComponentListener unifiedListener = new ComponentListener()
          {
            @Override
            public void modified( final Object source )
            {
              try
                {
                  item.setPattern( Pattern.compile( t.getText() ) );
                  sources.setExclusions( exclusions );
                  t.setBackground( normalBackground );
                  t.setToolTipText( null );
                }
              catch( final Exception x )
                {
                  t.setBackground( Color.yellow );
                  t.setToolTipText( "<html>" + "The regular expression is invalid:<br>" + x.getMessage() );
                }
            }
          };
          c.addActionListener( unifiedListener );
          c.addFocusListener( unifiedListener );
        }
      return c;
    }


    @Override
    public void addRow()
    {
      exclusions.add( new Exclusion() );
      sources.setExclusions( exclusions );
    }


    @Override
    public void delRow( final int rowIndex )
    {
      exclusions.remove( rowIndex );
      sources.setExclusions( exclusions );
    }


    @Override
    public Iterator<Exclusion> iterator()
    {
      return exclusions.iterator();
    }

    private final Sources sources;
    private final List<Exclusion> exclusions = new ArrayList<>();
    private final Map<Exclusion,JTextField> components = new HashMap<>();
  }


  private static class Renderer
    extends JLabel
    implements
      ListCellRenderer<FileHandlingOption>
  {

    public Renderer()
    {
      super();
      setOpaque( true );
      setBorder( BorderFactory.createEmptyBorder( 2,
                                                  4,
                                                  2,
                                                  0 ) );
    }


    @Override
    public Component getListCellRendererComponent( final JList<? extends FileHandlingOption> list,
                                                   final FileHandlingOption option,
                                                   final int index,
                                                   final boolean isSelected,
                                                   final boolean hasFocus )
    {
      setText( option.toString() + " \u2014 " + option.getToolTipText() );
      if( isSelected )
        {
          setForeground( list.getSelectionForeground() );
          setBackground( list.getSelectionBackground() );
        }
      else
        {
          setForeground( list.getForeground() );
          setBackground( Color.white );
        }
      return this;
    }

    private static final long serialVersionUID = -8534272763137518105L;
  }

  private static final Renderer renderer = new Renderer();
  private static final long serialVersionUID = 1044949688020881708L;
}
