package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;

import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Project.Binaries;
import com.ringlord.boxer2.Project.Classpath;
import com.ringlord.boxer2.Project.Library;
import com.ringlord.boxer2.Project.ProjectReference;
import com.ringlord.boxer2.Project.Sources;
import com.ringlord.swing.Images;
import com.ringlord.swing.TabTitle;


public class ProjectConfig
  extends JFrame
{
  public ProjectConfig( final Project project )
  {
    super( "Boxer - " + project.name() );

    setDefaultCloseOperation( JFrame.DO_NOTHING_ON_CLOSE );

    this.project = project;

    final JTabbedPane tabs = new JTabbedPane();

    final List<Binaries> binaries = project.getBinaries();
    final List<Classpath> classpath = project.getClasspath();
    final List<Library> libraries = project.getLibraries();
    final List<ProjectReference> projects = project.getProjects();
    final List<Sources> sources = project.getSources();

    final ConfigBinaries configBinaries = new ConfigBinaries( project );
    final ConfigClasspath configClasspath = new ConfigClasspath( project );
    final ConfigLibraries configLibraries = new ConfigLibraries( project );
    final ConfigProjectRefs configProjectRefs = new ConfigProjectRefs( project );
    final ConfigSources configSources = new ConfigSources( project );
    final ConfigAdditions configAdditions = new ConfigAdditions( project );
    final ConfigJNLP configJNLP = new ConfigJNLP( project );
    final ConfigMain configMain = new ConfigMain( project,
                                                  configJNLP );

    final List<Image> imageIcons = new ArrayList<>();
    imageIcons.add( Images.getImage( "/img/boxer2-96.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-64.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-48.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-32.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-24.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-16.png" ) );
    setIconImages( imageIcons );

    int tabIndex = 0;

    final JScrollPane mainScroll = new JScrollPane( configMain );
    mainScroll.getVerticalScrollBar().setUnitIncrement( 8 );
    TabTitle.addTab( tabs,
                     tabIndex++,
                     "Main",
                     Images.getIcon( "/img/house.png" ),
                     mainScroll,
                     false,
                     'M' );

    if( (binaries != null) && !binaries.isEmpty() )
      {
        final JScrollPane binariesScroll = new JScrollPane( configBinaries );
        binariesScroll.getVerticalScrollBar().setUnitIncrement( 8 );
        TabTitle.addTab( tabs,
                         tabIndex++,
                         "Binaries",
                         Images.getIcon( "/img/asterisk_orange.png" ),
                         binariesScroll,
                         false,
                         'B' );
      }
    if( (classpath != null) && !classpath.isEmpty() )
      {
        final JScrollPane classpathScroll = new JScrollPane( configClasspath );
        classpathScroll.getVerticalScrollBar().setUnitIncrement( 8 );
        TabTitle.addTab( tabs,
                         tabIndex++,
                         "Classpath",
                         Images.getIcon( "/img/basket.png" ),
                         classpathScroll,
                         false,
                         'H' );
      }
    if( (libraries != null) && !libraries.isEmpty() )
      {
        final JScrollPane librariesScroll = new JScrollPane( configLibraries );
        librariesScroll.getVerticalScrollBar().setUnitIncrement( 8 );
        TabTitle.addTab( tabs,
                         tabIndex++,
                         "Libraries",
                         Images.getIcon( "/img/bricks.png" ),
                         librariesScroll,
                         false,
                         'L' );
      }
    if( (projects != null) && !projects.isEmpty() )
      {
        final JScrollPane projectsScroll = new JScrollPane( configProjectRefs );
        projectsScroll.getVerticalScrollBar().setUnitIncrement( 8 );
        TabTitle.addTab( tabs,
                         tabIndex++,
                         "Projects",
                         Images.getIcon( "/img/application_view_tile.png" ),
                         projectsScroll,
                         false,
                         'R' );
      }
    if( (sources != null) && !sources.isEmpty() )
      {
        final JScrollPane sourcesScroll = new JScrollPane( configSources );
        sourcesScroll.getVerticalScrollBar().setUnitIncrement( 8 );
        TabTitle.addTab( tabs,
                         tabIndex++,
                         "Sources",
                         Images.getIcon( "/img/brick.png" ),
                         sourcesScroll,
                         false,
                         'S' );
      }

    final JScrollPane additionsScroll = new JScrollPane( configAdditions );
    additionsScroll.getVerticalScrollBar().setUnitIncrement( 8 );
    TabTitle.addTab( tabs,
                     tabIndex++,
                     "Additions",
                     Images.getIcon( "/img/page_add.png" ),
                     additionsScroll,
                     false,
                     'D' );

    // NOTE: The JNLP panel has a large scrollable area of its own
    tabIndex = tabs.getTabCount();
    TabTitle.addTab( tabs,
                     tabIndex++,
                     "JNLP",
                     Images.getIcon( "/img/cup.png" ),
                     configJNLP,
                     false,
                     'N' );

    final JButton about = new JButton( "About",
                                       Images.getIcon( "/img/information.png" ) );
    final JButton build = new JButton( "Package",
                                       Images.getIcon( "/img/package.png" ) );
    final JButton close = new JButton( "Close",
                                       Images.getIcon( "/img/cross.png" ) );

    about.setToolTipText( "Copyright, description, and license information" );
    build.setToolTipText( "Package the " + project.name() + " project" );
    close.setToolTipText( "Finish configuration" );

    about.setMnemonic( 'A' );
    build.setMnemonic( 'P' );
    close.setMnemonic( 'C' );

    final JFrame frame = this;

    addWindowListener( new WindowAdapter()
    {
      @Override
      public void windowClosing( final WindowEvent e )
      {
        save();
        dispose();
      }
    } );

    about.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        ProjectSelect.showAbout( frame,
                                 frame.getFont() );
      }
    } );

    build.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        final List<Project> one = new ArrayList<>();
        one.add( project );
        ProjectSelect.buildProjects( frame,
                                     one );
      }
    } );

    close.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        save();
        dispose();
      }
    } );

    final JPanel equalSize = new JPanel( new GridLayout( 1,
                                                         0,
                                                         4,
                                                         0 ) );
    equalSize.add( about );
    equalSize.add( build );
    equalSize.add( close );

    final JPanel cmdRow = new JPanel( new FlowLayout( FlowLayout.RIGHT,
                                                      0,
                                                      0 ) );
    cmdRow.add( equalSize );
    cmdRow.setBorder( BorderFactory.createEmptyBorder( 8,
                                                       0,
                                                       0,
                                                       0 ) );

    final JPanel layout = new JPanel( new BorderLayout() );
    layout.add( BorderLayout.CENTER,
                tabs );
    layout.add( BorderLayout.SOUTH,
                cmdRow );
    layout.setBorder( BorderFactory.createEmptyBorder( 12,
                                                       12,
                                                       12,
                                                       12 ) );

    setContentPane( layout );
    setSize( 800,
             600 );
  }


  public void save()
  {
    project.close();
  }

  private final Project project;
  private static final long serialVersionUID = -8011293391601597085L;
}
