package com.ringlord.boxer2.swing7;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dialog.ModalityType;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.DefaultListModel;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.ListCellRenderer;
import javax.swing.SwingWorker;
import javax.swing.border.Border;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import com.ringlord.Filename;
import com.ringlord.Settings;
import com.ringlord.boxer2.Builder;
import com.ringlord.boxer2.Manifest;
import com.ringlord.boxer2.Project;
import com.ringlord.boxer2.Util;
import com.ringlord.boxer2.Workspace;
import com.ringlord.swing.About;
import com.ringlord.swing.Images;


public class ProjectSelect
  extends JFrame
{
  public ProjectSelect()
  {
    super( "Boxer " + Util.VERSION );

    final JFrame frame = this;

    final JButton about = new JButton( "About",
                                       Images.getIcon( "/img/information.png" ) );
    final JButton configure = new JButton( "Configure",
                                           Images.getIcon( "/img/wrench.png" ) );
    final JButton build = new JButton( "Package",
                                       Images.getIcon( "/img/package.png" ) );
    final JButton exit = new JButton( "Exit",
                                      Images.getIcon( "/img/cross.png" ) );
    final JButton addWorkspace = new JButton( "Add\u2026",
                                              Images.getIcon( "/img/folder.png" ) );
    final JButton forgetWorkspace = new JButton( "Forget",
                                                 Images.getIcon( "/img/delete.png" ) );

    configure.setEnabled( false );
    build.setEnabled( false );
    forgetWorkspace.setEnabled( false );

    about.setToolTipText( "Copyright, description, and license information" );
    configure.setToolTipText( "Configure how the selected project is to be packaged" );
    build.setToolTipText( "Package the selected project(s)" );
    exit.setToolTipText( "Exit Boxer" );
    addWorkspace.setToolTipText( "Register a new Workspace" );
    forgetWorkspace.setToolTipText( "<html>"
                                    + "Unregister the selected workspace.<br>"
                                    + "<em>Don't fear, nothing is deleted,</em> you<br>"
                                    + "can always add/register it again." );

    about.setMnemonic( 'A' );
    configure.setMnemonic( 'C' );
    build.setMnemonic( 'P' );
    exit.setMnemonic( 'X' );
    addWorkspace.setMnemonic( 'D' );
    forgetWorkspace.setMnemonic( 'T' );

    final List<Image> imageIcons = new ArrayList<>();
    imageIcons.add( Images.getImage( "/img/boxer2-96.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-64.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-48.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-32.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-24.png" ) );
    imageIcons.add( Images.getImage( "/img/boxer2-16.png" ) );
    setIconImages( imageIcons );

    final int maxOpenConfigureAtOnce = Math.max( 1,
                                                 Util.getSettings().get( "configure-open-at-once-max",
                                                                         1 ) );

    // The listing of either one Workspace's projects, or all workspaces with
    // their projects indented
    final DefaultListModel<ComboEntry> entryModel = new DefaultListModel<>();
    final JList<ComboEntry> listing = new JList<>( entryModel );
    final ComboEntryRenderer comboEntryRenderer = new ComboEntryRenderer( listing );
    listing.setCellRenderer( comboEntryRenderer );
    listing.addListSelectionListener( new ListSelectionListener()
    {
      @Override
      public void valueChanged( final ListSelectionEvent e )
      {
        int subProjectCount = 0;
        final List<Project> projects = new ArrayList<>();
        for( final Object o : listing.getSelectedValuesList() )
          {
            final ComboEntry entry = (ComboEntry)o;
            if( entry.isProject() )
              {
                final Project p = entry.project();
                if( p.isSubproject() )
                  {
                    subProjectCount++;
                  }
                projects.add( p );
              }
          }

        configure.setEnabled( projects.size() == maxOpenConfigureAtOnce );

        build.setEnabled( (projects.size() - subProjectCount) > 0 );
      }
    } );
    listing.addMouseListener( new MouseAdapter()
    {
      @Override
      public void mouseClicked( final MouseEvent e )
      {
        if( e.getClickCount() == 2 ) // double click
          {
            final int[] selected = listing.getSelectedIndices();
            if( selected.length == 1 )
              {
                final ComboEntry entry = listing.getSelectedValue();
                if( entry.isProject() )
                  {
                    configureProject( frame,
                                      entry.project() );
                  }
              }
          }
        else if( e.isPopupTrigger() )
          {
            showPopupMenu( e );
          }
      }


      @Override
      public void mousePressed( final MouseEvent e )
      {
        if( e.isPopupTrigger() )
          {
            showPopupMenu( e );
          }
      }


      @Override
      public void mouseReleased( final MouseEvent e )
      {
        if( e.isPopupTrigger() )
          {
            showPopupMenu( e );
          }
      }


      private void showPopupMenu( final MouseEvent e )
      {
        final int index = listing.locationToIndex( e.getPoint() );
        if( index < 0 )
          {
            return;
          }
        listing.setSelectedIndex( index );
        final ComboEntry entry = listing.getSelectedValue();

        final Workspace workspace = (entry.isProject()
            ? null
            : entry.workspace());
        final Project project = (entry.isProject()
            ? entry.project()
            : null);

        final JMenuItem mRename = new JMenuItem( "Rename\u2026" );
        final JMenuItem mConfigure = new JMenuItem( "Configure\u2026" );
        final JMenuItem mBuild = new JMenuItem( "Build\u2026" );
        final JMenuItem mSetIcon = new JMenuItem( "Choose icon\u2026" );
        final JMenuItem mDelIcon = new JMenuItem( "Clear icon\u2026" );

        final JPopupMenu popup = new JPopupMenu();
        popup.add( mRename );
        popup.addSeparator();
        popup.add( mConfigure );
        popup.add( mBuild );
        popup.addSeparator();
        popup.add( mSetIcon );
        popup.add( mDelIcon );

        if( workspace != null )
          {
            mRename.addActionListener( new ActionListener()
            {
              @Override
              public void actionPerformed( final ActionEvent e )
              {
                final String value = JOptionPane.showInputDialog( ProjectSelect.this,
                                                                  "Rename workspace '" +
                                                                      workspace.getName() +
                                                                      "'\n" +
                                                                      "An empty value restores the original",
                                                                  workspace.getName() );
                if( value != null )
                  {
                    workspace.setName( value );
                    listing.repaint();
                  }
              }
            } );

            mConfigure.setEnabled( false );
            mBuild.setEnabled( false );
            mSetIcon.setEnabled( false );
            mDelIcon.setEnabled( false );
          }
        else
          {
            mRename.setEnabled( false );
            mRename.setToolTipText( "Not a workspace; rename projects using Eclipse" );

            mConfigure.addActionListener( new ActionListener()
            {
              @Override
              public void actionPerformed( final ActionEvent e )
              {
                configureProject( frame,
                                  project );
              }
            } );

            if( project.isSubproject() )
              {
                mBuild.setEnabled( false );
                mBuild.setToolTipText( "Marked as a sub-project, not buildable by itself" );
              }
            else
              {
                mBuild.addActionListener( new ActionListener()
                {
                  @Override
                  public void actionPerformed( final ActionEvent e )
                  {
                    project.save();

                    final List<Project> projects = new ArrayList<>();
                    projects.add( project );

                    buildProjects( frame,
                                   projects );
                  }
                } );
              }

            mSetIcon.addActionListener( new ActionListener()
            {
              @Override
              public void actionPerformed( final ActionEvent e )
              {
                if( fileChooser == null )
                  {
                    fileChooser = new JFileChooser();
                  }
                fileChooser.setCurrentDirectory( project.root() );
                fileChooser.setDialogTitle( "Select Icon for '" + project.name() + "'" );
                if( fileChooser.showOpenDialog( listing ) == JFileChooser.APPROVE_OPTION )
                  {
                    final File picked = fileChooser.getSelectedFile();
                    try( final InputStream f = new FileInputStream( picked ) )
                      {
                        final BufferedImage bufferedImage = ImageIO.read( f );
                        project.setIcon( new ImageIcon( bufferedImage ) );
                        project.save();
                        listing.repaint();
                      }
                    catch( final IOException x )
                      {
                      }
                  }
              }
            } );

            if( project.getIcon() == null )
              {
                mDelIcon.setEnabled( false );
                mDelIcon.setToolTipText( "No project icon to remove" );
              }
            else
              {
                mDelIcon.addActionListener( new ActionListener()
                {
                  @Override
                  public void actionPerformed( final ActionEvent e )
                  {
                    project.setIcon( null );
                    project.save();
                    listing.repaint();
                  }
                } );
              }
          }

        popup.show( listing,
                    e.getX() - 2,
                    e.getY() - 2 );

      }

      private JFileChooser fileChooser = null;
    } );

    // The dropdown selector of either ALL or a single Workspace
    final DefaultComboBoxModel<WorkspaceEntry> workspaceModel = new DefaultComboBoxModel<>();
    workspaceModel.addElement( new WorkspaceEntry() );
    final List<Workspace> knownWorkspaces = Util.getKnownWorkspaces();
    for( final Workspace w : knownWorkspaces )
      {
        workspaceModel.addElement( new WorkspaceEntry( w ) );
      }
    final JComboBox<WorkspaceEntry> workspaceSelector = new JComboBox<>( workspaceModel );
    workspaceSelector.setRenderer( new WorkspaceEntryRenderer() );
    workspaceSelector.setToolTipText( "Projects in one workspace, or in all registered workspaces" );
    workspaceSelector.addItemListener( new ItemListener()
    {
      @Override
      public void itemStateChanged( final ItemEvent e )
      {
        if( e.getStateChange() == ItemEvent.SELECTED )
          {
            final WorkspaceEntry we = (WorkspaceEntry)workspaceSelector.getSelectedItem();
            if( we == null )
              {
                forgetWorkspace.setEnabled( false );
                entryModel.clear();
                Util.setCurrentWorkspace( null );
              }
            else
              {
                final Workspace w = we.workspace();
                if( w == null ) // sanity check
                  {
                    forgetWorkspace.setEnabled( false );
                    loadModel( Util.getKnownWorkspaces(),
                               comboEntryRenderer,
                               entryModel );
                    Util.setCurrentWorkspace( null );
                  }
                else
                  {
                    forgetWorkspace.setEnabled( true );
                    loadModel( w,
                               comboEntryRenderer,
                               entryModel );
                    Util.setCurrentWorkspace( w.getName() );
                  }
              }
          }
      }
    } );

    boolean loadedWorkspace = false;
    final String savedWorkspace = Util.getCurrentWorkspace();
    if( savedWorkspace != null )
      {
        final int count = workspaceModel.getSize();
        for( int i = 0; i < count; i++ )
          {
            final WorkspaceEntry we = workspaceModel.getElementAt( i );
            final Workspace w = we.workspace();
            if( w != null )
              {
                if( savedWorkspace.equals( w.getName() ) )
                  {
                    workspaceSelector.setSelectedItem( we );
                    forgetWorkspace.setEnabled( true );
                    loadModel( w,
                               comboEntryRenderer,
                               entryModel );
                    loadedWorkspace = true;
                    break;
                  }
              }
          }
      }
    if( !loadedWorkspace )
      {
        loadModel( knownWorkspaces,
                   comboEntryRenderer,
                   entryModel );
      }

    addWorkspace.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        if( fileChooser == null )
          {
            fileChooser = new JFileChooser( System.getProperty( "user.home" ) );
            fileChooser.setFileSelectionMode( JFileChooser.DIRECTORIES_ONLY );
            fileChooser.setDialogType( JFileChooser.OPEN_DIALOG );
            fileChooser.setDialogTitle( "Select Eclipse Workspace" );
          }
        final int choice = fileChooser.showOpenDialog( ProjectSelect.this );
        if( choice == 0 )
          {
            try
              {
                final Workspace w = new Workspace( fileChooser.getSelectedFile() );
                final WorkspaceEntry wEntry = new WorkspaceEntry( w );

                final List<Workspace> all = Util.getKnownWorkspaces();
                if( !all.contains( w ) )
                  {
                    all.add( w );
                    Util.setKnownWorkspaces( all );

                    workspaceModel.addElement( wEntry );
                    workspaceModel.setSelectedItem( wEntry );
                  }
              }
            catch( ParserConfigurationException |
                   SAXException |
                   IOException x )
              {
                // TODO Auto-generated catch block
                x.printStackTrace();
              }
          }
      }
      private JFileChooser fileChooser;
    } );

    forgetWorkspace.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        final WorkspaceEntry we = (WorkspaceEntry)workspaceSelector.getSelectedItem();
        if( we != null )
          {
            final List<Workspace> all = Util.getKnownWorkspaces();
            all.remove( we.workspace() );
            Util.setKnownWorkspaces( all );

            workspaceSelector.removeItem( we );
          }
      }
    } );

    final JPanel equalSize = new JPanel( new GridLayout( 1,
                                                         0,
                                                         4,
                                                         0 ) );
    equalSize.add( about );
    equalSize.add( configure );
    equalSize.add( build );
    equalSize.add( exit );

    final JPanel commandRow = new JPanel( new FlowLayout( FlowLayout.RIGHT,
                                                          0,
                                                          0 ) );
    commandRow.add( equalSize );

    final JPanel listingPanel = new JPanel( new BorderLayout() );
    listingPanel.add( BorderLayout.CENTER,
                      new JScrollPane( listing ) );
    listingPanel.setBorder( BorderFactory.createEmptyBorder( 8,
                                                             0,
                                                             8,
                                                             0 ) );

    final JPanel workspaceAddForget = new JPanel( new GridLayout( 1,
                                                                  0,
                                                                  0,
                                                                  0 ) );
    workspaceAddForget.add( addWorkspace );
    workspaceAddForget.add( forgetWorkspace );
    workspaceAddForget.setBorder( BorderFactory.createEmptyBorder( 0,
                                                                   8,
                                                                   0,
                                                                   0 ) );
    final JPanel selectorRow = new JPanel( new BorderLayout() );
    selectorRow.add( BorderLayout.CENTER,
                     workspaceSelector );
    selectorRow.add( BorderLayout.EAST,
                     workspaceAddForget );

    final JPanel layout = new JPanel( new BorderLayout() );
    layout.add( BorderLayout.NORTH,
                selectorRow );
    layout.add( BorderLayout.CENTER,
                listingPanel );
    layout.add( BorderLayout.SOUTH,
                commandRow );
    layout.setBorder( BorderFactory.createEmptyBorder( 12,
                                                       12,
                                                       12,
                                                       12 ) );

    about.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        showAbout( frame,
                   listing.getFont() );
      }
    } );

    configure.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        final ComboEntry entry = listing.getSelectedValue();
        if( entry.isProject() )
          {
            final Project p = entry.project();
            configureProject( frame,
                              p );
          }
      }
    } );

    build.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        final List<Project> projects = new ArrayList<>();
        for( final Object o : listing.getSelectedValuesList() )
          {
            final ComboEntry entry = (ComboEntry)o;
            if( entry.isProject() )
              {
                projects.add( entry.project() );
              }
          }
        if( !projects.isEmpty() )
          {
            for( final Project p : projects )
              {
                p.save();
              }
            buildProjects( frame,
                           projects );
          }
      }
    } );

    exit.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent e )
      {
        shutdown();
      }
    } );

    addWindowListener( new WindowAdapter()
    {
      @Override
      public void windowClosing( final WindowEvent e )
      {
        shutdown();
      }
    } );

    setDefaultCloseOperation( JFrame.DO_NOTHING_ON_CLOSE );
    setContentPane( layout );
    pack();

    final Rectangle bounds = frame.getBounds();
    bounds.width = Math.max( bounds.width,
                             400 );
    bounds.height = Math.max( bounds.width,
                              600 );
    frame.setBounds( Util.getSettings().get( "main-window-xywh",
                                             bounds ) );
  }


  public void shutdown()
  {
    setVisible( false );
    Util.getSettings().put( "main-window-xywh",
                            getBounds() );
    final Settings settings = Util.getSettings();
    boolean isFirst = true;
    for( final ProjectConfig p : open.values() )
      {
        p.setVisible( false );
        p.save();
        if( isFirst )
          {
            settings.put( "project-window-xywh",
                          p.getBounds() );
            isFirst = false;
          }
        p.dispose();
      }
    dispose();
    System.exit( 0 ); // not strictly necessary but faster
  }


  private void configureProject( final JFrame frame,
                                 final Project project )
  {
    final ProjectConfig w = open.get( project );
    if( w == null )
      {
        final ProjectConfig c = new ProjectConfig( project );
        c.addWindowListener( new WindowAdapter()
        {
          @Override
          public void windowClosed( final WindowEvent e )
          {
            open.remove( project );
            c.removeWindowListener( this );

            Util.getSettings().put( "project-window-xywh",
                                    c.getBounds() );

            final int extendedState = frame.getExtendedState();
            if( (extendedState & JFrame.ICONIFIED) == JFrame.ICONIFIED )
              {
                frame.setExtendedState( extendedState - JFrame.ICONIFIED );
              }
            repaint();
            frame.setVisible( true );
            frame.toFront();
          }
        } );
        open.put( project,
                  c );

        final Settings settings = Util.getSettings();
        Rectangle bounds = settings.get( "project-window-xywh",
                                         (Rectangle)null );
        if( bounds == null )
          {
            c.pack();
            final Rectangle parent = ProjectSelect.this.getBounds();
            bounds = c.getBounds();
            bounds.x = parent.x + 32;
            bounds.y = parent.y + 32;
            bounds.width = Math.max( bounds.width,
                                     800 );
            bounds.height = Math.max( bounds.height,
                                      600 );
          }
        c.setBounds( bounds );
        c.setVisible( true );
      }
    else
      {
        final int extendedState = w.getExtendedState();
        if( (extendedState & JFrame.ICONIFIED) == JFrame.ICONIFIED )
          {
            w.setExtendedState( extendedState - JFrame.ICONIFIED );
          }
        w.setVisible( true );
        w.toFront();
      }
  }


  static void showAbout( final JFrame parent,
                         final Font font )
  {
    final JDialog d = new JDialog( parent,
                                   "About Boxer",
                                   ModalityType.DOCUMENT_MODAL );
    final About about = new com.ringlord.swing.About( ABOUT_HTML_BODY,
                                                      font,
                                                      Images.getIcon( "/img/boxer2-128.png" ),
                                                      true );
    about.registerKeyboardAction( new ActionListener()
                                  {
                                    @Override
                                    public void actionPerformed( final ActionEvent e )
                                    {
                                      d.setVisible( false );
                                    }
                                  },
                                  KeyStroke.getKeyStroke( KeyEvent.VK_ESCAPE,
                                                          0 ),
                                  JComponent.WHEN_IN_FOCUSED_WINDOW );
    d.setContentPane( about );
    d.setSize( 640,
               500 );
    d.setLocationRelativeTo( parent );
    d.setVisible( true );
  }


  /**
   * All Entry elements refer to Projects within a single selected Workspace
   *
   * @param workspace
   * @param listing
   * @param model
   */
  private void loadModel( final Workspace workspace,
                          final ComboEntryRenderer renderer,
                          final DefaultListModel<ComboEntry> model )
  {
    model.clear();
    for( final Project p : workspace.projects() )
      {
        model.addElement( new ComboEntry( p ) );
      }
    renderer.setMixedMode( false );
  }


  /**
   * Entry elements refer to Workspace as a header and Projects within the
   * previously listed Workspace.
   *
   * @param workspaces
   * @param listing
   * @param model
   */
  private void loadModel( final List<Workspace> workspaces,
                          final ComboEntryRenderer renderer,
                          final DefaultListModel<ComboEntry> model )
  {
    model.clear();
    for( final Workspace w : workspaces )
      {
        model.addElement( new ComboEntry( w ) );
        for( final Project p : w.projects() )
          {
            model.addElement( new ComboEntry( p ) );
          }
      }
    renderer.setMixedMode( true );
  }


  static void buildProjects( final JFrame frame,
                             final List<Project> projects )
  {
    final JTextArea output = new JTextArea();
    output.setEditable( false );
    output.setFont( new Font( "Monospaced",
                              Font.PLAIN,
                              output.getFont().getSize() ) );

    final JButton close = new JButton( "Close",
                                       Images.getIcon( "/img/cross.png" ) );
    close.setToolTipText( "Close the window. You can also press ESCAPE" );
    close.setEnabled( false );

    final JPanel cmdRow = new JPanel( new FlowLayout( FlowLayout.RIGHT,
                                                      0,
                                                      0 ) );
    cmdRow.add( close );
    cmdRow.setBorder( BorderFactory.createEmptyBorder( 8,
                                                       0,
                                                       0,
                                                       0 ) );

    final JPanel layout = new JPanel( new BorderLayout() );
    layout.add( BorderLayout.CENTER,
                new JScrollPane( output ) );
    layout.add( BorderLayout.SOUTH,
                cmdRow );
    layout.setBorder( BorderFactory.createEmptyBorder( 12,
                                                       12,
                                                       12,
                                                       12 ) );

    final JDialog d = new JDialog( frame,
                                   "Build Results",
                                   ModalityType.DOCUMENT_MODAL );
    layout.registerKeyboardAction( new ActionListener()
                                   {
                                     @Override
                                     public void actionPerformed( final ActionEvent arg0 )
                                     {
                                       d.setVisible( false );
                                     }
                                   },
                                   KeyStroke.getKeyStroke( KeyEvent.VK_ESCAPE,
                                                           0 ),
                                   JComponent.WHEN_IN_FOCUSED_WINDOW );
    d.setContentPane( layout );
    d.setSize( 640,
               480 );
    d.setLocationRelativeTo( frame );
    d.setBounds( Util.getSettings().get( "build-window-xywh",
                                         d.getBounds() ) );

    close.addActionListener( new ActionListener()
    {
      @Override
      public void actionPerformed( final ActionEvent arg0 )
      {
        d.setVisible( false );
      }
    } );

    new SwingWorker<Void,Void>()
    {
      @Override
      public Void doInBackground()
      {
        buildProjects( projects,
                       output );
        close.setEnabled( true );
        return null;
      }
    }.execute();

    d.setVisible( true );
    Util.getSettings().put( "build-window-xywh",
                            d.getBounds() );
  }


  /**
   *
   * @param projects
   * @param output
   *          null if output is to go to the console
   */
  private static void buildProjects( final List<Project> projects,
                                     final JTextArea output )
  {
    final Logger logger = Util.logger();
    final long startTimeAll = System.currentTimeMillis();
    int processed = 0;
    boolean isFirst = true;
    for( final Project project : projects )
      {
        project.save();

        final long startTime = System.currentTimeMillis();

        if( isFirst )
          {
            isFirst = false;
          }
        else
          {
            if( output != null )
              {
                output.append( "\n" );
              }
          }

        if( project.isSubproject() )
          {
            final String msg = "Skipping " + project + " (Marked as a subproject, not buildable by itself)";
            logger.log( Level.WARNING,
                        msg );
            if( output != null )
              {
                output.append( msg + "\n" );
              }
            continue;
          }

        final long buildNumber = project.bumpBuildNumber();
        logger.log( Level.INFO,
                    "Processing " + project + " (Build " + buildNumber + ")" );
        if( output != null )
          {
            output.append( "Processing " + project + " (Build " + buildNumber + ")\n" );
          }

        final File archive = Util.resolveFilename( project,
                                                   project.getOutputJarFile() );
        final File distro = Util.resolveFilename( project,
                                                  project.getDistributionFile() );
        final File workingPath = Filename.resolveTilde( project.getWorkingDirectory() );

        try
          {
            // Attached to the log, copies certain records to the 'output'
            final Handler logFilter = new Handler()
            {
              @Override
              public void close()
                throws SecurityException
              {
              }


              @Override
              public void flush()
              {
              }


              @Override
              public void publish( final LogRecord record )
              {
                final Level level = record.getLevel();
                if( level == Level.SEVERE )
                  {
                    output.append( "ERROR: " + record.getMessage() + "\n" );
                  }
                else if( level == Level.WARNING )
                  {
                    output.append( "WARNING: " + record.getMessage() + "\n" );
                  }
                else if( level == Level.CONFIG )
                  {
                    output.append( ">>> " + record.getMessage() + "\n" );
                  }
                else if( (level == Level.INFO) || (level == Level.ALL) )
                  {
                    output.append( record.getMessage() + "\n" );
                  }
              }
            };
            Util.logger().addHandler( logFilter );

            try
              {
                // External library file references that may need to be package
                // into the distribution archive
                final List<Project.Library> externalLibs = new ArrayList<>();
                final Manifest manifest = Builder.assemble( project,
                                                            workingPath,
                                                            externalLibs );

                Builder.writeJAR( project,
                                  archive,
                                  manifest );
                Builder.signJAR( project,
                                 archive,
                                 new PasswordReaderGUI( output ) );
                try
                  {
                    archive.setExecutable( true );
                  }
                catch( final Throwable x )
                  {
                    x.printStackTrace();
                  }
                Builder.writeJNLP( project,
                                   archive );
                if( project.createDistributionFile() )
                  {
                    Builder.writeDistro( distro,
                                         archive,
                                         Util.resolveFilename( project,
                                                               project.getDistributionSubdir() ).toString(),
                                         externalLibs,
                                         project.getAdditions() );
                    if( output != null )
                      {
                        output.append( "Wrote " + distro + "\n" );
                      }
                  }
                Builder.writeDigestFiles( project,
                                          archive,
                                          (project.createDistributionFile()
                                              ? distro
                                              : null) );
                project.save();
              }
            finally
              {
                Util.logger().removeHandler( logFilter );
              }
          }
        catch( NoSuchAlgorithmException |
               IOException e )
          {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
        final long totalTime = System.currentTimeMillis() - startTime;
        final String result = String.format( "Completed project " + project + " in %1.3f seconds",
                                             totalTime / 1000.0d );
        Util.logger().log( Level.INFO,
                           result );
        if( output != null )
          {
            output.append( result + "\n" );
          }
        processed++;
      }

    if( processed > 1 )
      {
        final long totalTimeAll = System.currentTimeMillis() - startTimeAll;
        final String result = String.format( "Completed all %d projects in %1.3f seconds",
                                             projects.size(),
                                             totalTimeAll / 1000.0d );
        Util.logger().log( Level.INFO,
                           result );
        if( output != null )
          {
            output.append( "\n" + result + "\n" );
          }
      }
  }


  /**
   * Renders the entries in the project selection listing in two different
   * modes, where workspace names are rendered in a different style than project
   * names.
   *
   * @author K Udo Schuermann
   */
  private class ComboEntryRenderer
    extends JLabel
    implements
      ListCellRenderer<ComboEntry>
  {
    ComboEntryRenderer( final JList<ComboEntry> list )
    {
      super();

      final float size = list.getFont().getSize2D();
      this.workspaceFont = list.getFont().deriveFont( Font.BOLD ).deriveFont( size * 1.1f );
      this.workspaceBorder = BorderFactory.createEmptyBorder( 3,
                                                              4,
                                                              3,
                                                              0 );

      this.projectFont = list.getFont().deriveFont( Font.PLAIN );
      this.subProjFont = list.getFont().deriveFont( Font.ITALIC );
      this.projectFontOpen = list.getFont().deriveFont( Font.BOLD );
      this.subProjFontOpen = list.getFont().deriveFont( Font.ITALIC | Font.BOLD );
      this.projectBorderMixed = BorderFactory.createEmptyBorder( 2,
                                                                 24,
                                                                 2,
                                                                 0 );
      this.projectBorderSingle = BorderFactory.createEmptyBorder( 2,
                                                                  4,
                                                                  2,
                                                                  0 );
    }


    void setMixedMode( final boolean isMixedMode )
    {
      this.isMixedMode = isMixedMode;
    }


    @Override
    public Component getListCellRendererComponent( final JList<? extends ComboEntry> list,
                                                   final ComboEntry entry,
                                                   final int index,
                                                   final boolean isSelected,
                                                   final boolean hasFocus )
    {
      if( entry.isProject() )
        {
          final Project project = entry.project();
          if( project.isSubproject() )
            {
              if( open.containsKey( project ) )
                {
                  setFont( subProjFontOpen );
                }
              else
                {
                  setFont( subProjFont );
                }
              setToolTipText( "<html>" +
                              "Packaged as part of one or more<br>" +
                              "other projects and not by itself.<br>" +
                              project.root().toString() );
            }
          else
            {
              if( open.containsKey( project ) )
                {
                  setFont( projectFontOpen );
                }
              else
                {
                  setFont( projectFont );
                }
              setToolTipText( project.root().toString() );
            }
          setText( project.name() );
          setIcon( project.getIcon() );
          if( isMixedMode )
            {
              setBorder( projectBorderMixed );
            }
          else
            {
              setBorder( projectBorderSingle );
            }
        }
      else
        // it is a Workspace
        {
          final Workspace workspace = entry.workspace();
          setFont( workspaceFont );
          setText( workspace.getName() );
          setIcon( null );
          setToolTipText( "<html>" + "Workspace for projects:<br>" + workspace.root() );
          setBorder( workspaceBorder );
        }

      if( isSelected )
        {
          setOpaque( true );
          setForeground( list.getSelectionForeground() );
          setBackground( list.getSelectionBackground() );
        }
      else
        {
          setOpaque( false );
          setForeground( list.getForeground() );
          setBackground( list.getBackground() );
        }

      return this;
    }
    private boolean isMixedMode;
    private final Font workspaceFont, projectFont, subProjFont, projectFontOpen, subProjFontOpen;
    private final Border workspaceBorder, projectBorderMixed, projectBorderSingle;
    private static final long serialVersionUID = -4986798575957253247L;
  }


  private class ComboEntry
  {
    ComboEntry( final Workspace workspace )
    {
      super();
      this.workspace = workspace;
      this.project = null;
    }


    ComboEntry( final Project project )
    {
      super();
      this.workspace = null;
      this.project = project;
    }


    boolean isProject()
    {
      return (project != null);
    }


    Workspace workspace()
    {
      return workspace;
    }


    Project project()
    {
      return project;
    }

    private final Workspace workspace;
    private final Project project;
  }


  private class WorkspaceEntry
  {
    WorkspaceEntry()
    {
      this.workspace = null;
    }


    WorkspaceEntry( final Workspace workspace )
    {
      this.workspace = workspace;
    }


    Workspace workspace()
    {
      return workspace;
    }
    private final Workspace workspace;
  }


  /**
   * Renders the entries for the ComboBox
   *
   * @author K Udo Schuermann
   */
  private class WorkspaceEntryRenderer
    extends JLabel
    implements
      ListCellRenderer<WorkspaceEntry>
  {
    WorkspaceEntryRenderer()
    {
      setBorder( BorderFactory.createEmptyBorder( 3,
                                                  8,
                                                  3,
                                                  0 ) );
      specialFont = getFont().deriveFont( Font.ITALIC );
      normalFont = getFont().deriveFont( Font.PLAIN );
    }


    @Override
    public Component getListCellRendererComponent( final JList<? extends WorkspaceEntry> list,
                                                   final WorkspaceEntry workspace,
                                                   final int index,
                                                   final boolean isSelected,
                                                   final boolean hasFocus )
    {
      final Workspace w = workspace.workspace();
      if( w == null )
        {
          setFont( specialFont );
          setText( "All Workspaces" );
          setToolTipText( null );
        }
      else
        {
          setFont( normalFont );
          setText( w.getName() );
          setToolTipText( w.root().toString() );
        }

      if( isSelected )
        {
          setOpaque( true );
          setForeground( list.getSelectionForeground() );
          setBackground( list.getSelectionBackground() );
        }
      else
        {
          setOpaque( false );
          setForeground( list.getForeground() );
          setBackground( list.getBackground() );
        }
      return this;
    }

    private final Font normalFont, specialFont;
    private static final long serialVersionUID = 5180572958376992560L;
  }

  private final Map<Project,ProjectConfig> open = new HashMap<>();
  private static final String ABOUT_HTML_BODY = "<b>Boxer</b> " +
                                                Util.VERSION +
                                                " (" +
                                                Util.VERDATE +
                                                ")<br>\n" +
                                                "Copyright \u00a9 " +
                                                Util.CPYEARS +
                                                " by K. Udo Schuermann<br>\n" +
                                                "All rights reserved<br>" +
                                                "<p>Uses Eclipse .classpath information to assist in packaging Java " +
                                                "software for distribution. For more information, updates, and " +
                                                "documentation, please visit " +
                                                "<a href=\"http://ringlord.com/boxer.html\">http://ringlord.com/boxer.html</a></p>" +

                                                "<p>My thanks also go to Mark James for his awesome Silk icon set: " +
                                                "<a href=\"http://www.famfamfam.com/lab/icons/silk/\">" +
                                                "http://www.famfamfam.com/lab/icons/silk/</a></p>" +

                                                "<p><b>License</b> (GNU General Public License v3)</p>" +
                                                "<p>This program is free software: you can redistribute it and/or modify " +
                                                "it under the terms of the GNU General Public License as published by " +
                                                "the Free Software Foundation, either version 3 of the License, or " +
                                                "(at your option) any later version.</p>" +
                                                "<p>This program is distributed in the hope that it will be useful, " +
                                                "but WITHOUT ANY WARRANTY; without even the implied warranty of " +
                                                "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the " +
                                                "GNU General Public License for more details.</p>" +
                                                "<p>A copy of the GNU General Public License should be included with " +
                                                "this program (inside the JAR and/or as a separate file). If not, see " +
                                                "<a href=\"http://www.gnu.org/licenses/\">http://www.gnu.org/licenses/</a></p>";
  private static final long serialVersionUID = 4434901772620351660L;
}
